from __future__ import annotations

import abc
import argparse
import asyncio
import dataclasses
import math
import os
import pickle
import platform
import shlex
import time
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

import cloudpickle
import kubernetes_asyncio.client as kubernetes_client
import kubernetes_asyncio.client.exceptions as kubernetes_client_exceptions
import kubernetes_asyncio.config as kubernetes_config
import kubernetes_asyncio.stream as kubernetes_stream

import meadowrun.func_worker_storage_helper
from meadowrun.config import GPU, LOGICAL_CPU, MEMORY_GB, MEADOWRUN_INTERPRETER
from meadowrun.credentials import KubernetesSecretRaw
from meadowrun.docker_controller import expand_ports
from meadowrun.k8s_integration.k8s_core import (
    get_main_container_is_ready,
    get_pods_for_job,
    run_command_on_pod,
    set_main_container_ready,
    wait_for_pod,
    wait_for_pod_running,
)
from meadowrun.meadowrun_pb2 import (
    Job,
    ProcessState,
    PyAgentJob,
    ServerAvailableInterpreter,
)
from meadowrun.run_job_core import (
    Host,
    JobCompletion,
    MeadowrunException,
    TaskResult,
    WaitOption,
    _PRINT_RECEIVED_TASKS_SECONDS,
)
from meadowrun.run_job_local import (
    TaskWorkerServer,
    WorkerMonitor,
    _get_credentials_for_docker,
    _get_credentials_sources,
    _string_pairs_to_dict,
    restart_worker,
)
from meadowrun.storage_grid_job import (
    complete_task,
    download_task_arg,
    get_job_completion_from_process_state,
    receive_results,
    upload_task_args,
)
from meadowrun.storage_keys import (
    storage_key_job_to_run,
    storage_key_ranges,
    storage_key_task_args,
)
from meadowrun.version import __version__

if TYPE_CHECKING:
    from asyncio import Task

    from meadowrun.abstract_storage_bucket import AbstractStorageBucket
    from meadowrun.instance_selection import ResourcesInternal
    from meadowrun.run_job_core import WorkerProcessState, TaskProcessState

MEADOWRUN_POD_NAME = "MEADOWRUN_POD_NAME"


_T = TypeVar("_T")
_U = TypeVar("_U")


async def _indexed_map_worker(
    total_num_tasks: int,
    num_workers: int,
    job_id: str,
    worker_server: TaskWorkerServer,
    worker_monitor: WorkerMonitor,
) -> None:
    """
    This is a worker function to help with running a run_map. This worker assumes that
    JOB_COMPLETION_INDEX is set, which Kubernetes will set for indexed completion jobs.
    This worker assumes task arguments are accessible via
    meadowrun.func_worker_storage_helper.FUNC_WORKER_STORAGE_BUCKET and will just
    complete all of the tasks where task_index % num_workers == current worker index.
    """

    # WORKER_INDEX will be available in reusable pods. In non-reusable pods we have to
    # use JOB_COMPLETION_INDEX
    current_worker_index = int(
        os.environ.get("MEADOWRUN_WORKER_INDEX", os.environ["JOB_COMPLETION_INDEX"])
    )

    # we're always being called from run_job_local_storage_main which sets these
    # variables for us
    storage_bucket = meadowrun.func_worker_storage_helper.FUNC_WORKER_STORAGE_BUCKET
    if storage_bucket is None:
        raise ValueError(
            "Programming error--_indexed_map_worker must be called from "
            "run_job_local_storage_main"
        )

    # only the first option (MEADOWRUN_POD_NAME) will give the actual name of the pod.
    # The hostname is similar but slightly different, and we'll fall back on that if
    # something has gone wrong and MEADOWRUN_POD_NAME is not available
    pod_name = os.environ.get(
        MEADOWRUN_POD_NAME, os.environ.get("HOSTNAME", platform.node())
    )
    log_file_name = f"{pod_name}:/var/meadowrun/job_logs/{job_id}.log"

    byte_ranges = pickle.loads(
        await storage_bucket.get_bytes(storage_key_ranges(job_id))
    )

    i = current_worker_index
    while i < total_num_tasks:
        arg = await download_task_arg(storage_bucket, job_id, byte_ranges[i])

        try:
            worker_monitor.start_stats()
            await worker_server.send_message(arg)
            state, result_bytes = await worker_server.receive_message()
            stats = await worker_monitor.stop_stats()

            process_state = ProcessState(
                state=ProcessState.ProcessStateEnum.SUCCEEDED
                if state == "SUCCEEDED"
                else ProcessState.ProcessStateEnum.PYTHON_EXCEPTION,
                pickled_result=result_bytes,
                return_code=0,
                max_memory_used_gb=stats.max_memory_used_gb,
                log_file_name=log_file_name,
            )
        except BaseException:
            stats = await worker_monitor.stop_stats()
            process_state = ProcessState(
                state=ProcessState.ProcessStateEnum.UNEXPECTED_WORKER_EXIT,
                return_code=(await worker_monitor.try_get_return_code()) or 0,
                max_memory_used_gb=stats.max_memory_used_gb,
                log_file_name=log_file_name,
            )
            await restart_worker(worker_server, worker_monitor)

        # we don't support retries yet so we're always on attempt 1
        await complete_task(storage_bucket, job_id, i, 1, process_state)

        print(
            f"Meadowrun agent: Completed task #{i}, "
            f"state {ProcessState.ProcessStateEnum.Name(process_state.state)}, max "
            f"memory {process_state.max_memory_used_gb}GB "
        )

        i += num_workers


async def _image_name_from_job(job: Job) -> Tuple[bool, str, Optional[str], Job]:
    """
    Returns is_custom_container_image, image_name, image_pull_secret_name, and
    modified_job. modified_job should NOT be used with run_direct_command--it has been
    modified to run inside of a custom container.
    """

    # get the image_repository_name and image_name
    interpreter_deployment_type = job.WhichOneof("interpreter_deployment")

    if interpreter_deployment_type == "container_at_digest":
        is_custom_container_image = True
        image_repository_name = job.container_at_digest.repository
        image_name = (
            f"{job.container_at_digest.repository}@" f"{job.container_at_digest.digest}"
        )
    elif interpreter_deployment_type == "container_at_tag":
        is_custom_container_image = True
        image_repository_name = job.container_at_tag.repository
        image_name = f"{job.container_at_tag.repository}:{job.container_at_tag.tag}"
    elif interpreter_deployment_type == "server_available_container":
        is_custom_container_image = True
        image_repository_name = None
        image_name = f"{job.server_available_container.image_name}"
    else:
        is_custom_container_image = False
        image_repository_name = None
        # This else block implies that the user specified an environment spec rather
        # than a container. Our approach is to default to a "generic container image"
        # and build the environment inside of the container. We could imagine other
        # approaches like building the container locally and uploading it to Kubernetes.

        if interpreter_deployment_type == "environment_spec_in_code":
            python_version = job.environment_spec_in_code.python_version
        elif interpreter_deployment_type == "environment_spec":
            python_version = job.environment_spec.python_version
        elif interpreter_deployment_type == "server_available_interpreter":
            python_version = None
        else:
            raise ValueError(
                "Programming error: unknown interpreter deployment type "
                f"{interpreter_deployment_type}"
            )

        if not python_version:
            # conda environments and raw server_available_interpreter won't have a
            # python version
            python_version = "3.10"

        # TODO use meadowrun-cuda if we need cuda
        image_name = f"meadowrun/meadowrun:{__version__}-py{python_version}"

    # next, get an image pull secrets if it's been provided
    image_pull_secret_name = None
    if image_repository_name is not None:
        image_pull_secret = await _get_credentials_for_docker(
            image_repository_name, _get_credentials_sources(job), None
        )
        if image_pull_secret is not None:
            if not isinstance(image_pull_secret, KubernetesSecretRaw):
                raise NotImplementedError(
                    "Using anything other than KubernetesSecret with a "
                    "Kubernetes host has not been implemented, "
                    f"{type(image_pull_secret)} was provided"
                )
            image_pull_secret_name = image_pull_secret.secret_name

    if is_custom_container_image:
        # if we have a custom container and we're going to run a job on it via
        # run_job_local_storage_main, we need to tell run_job_local_storage_main to just
        # use the container's interpreter (rather than trying to go fetch another
        # container)

        # it's a little paranoid to make a copy here, this could probably be optimized
        # if we need to for performance
        modified_job = Job()
        modified_job.CopyFrom(job)
        # MEADOWRUN_INTERPRETER in this case will refer to whatever interpreter is
        # running Meadowrun which will be whatever interpreter is on the path in the
        # custom container
        modified_job.server_available_interpreter.CopyFrom(
            ServerAvailableInterpreter(interpreter_path=MEADOWRUN_INTERPRETER)
        )
    else:
        modified_job = job

    return is_custom_container_image, image_name, image_pull_secret_name, modified_job


class StorageBucketSpec(abc.ABC):
    """
    An abstract class that specifies an object storage system that Meadowrun can use to
    send data back and forth from remote workers
    """

    @abc.abstractmethod
    async def get_storage_bucket(
        self, kubernetes_namespace: str
    ) -> AbstractStorageBucket:
        ...

    @abc.abstractmethod
    def get_command_line_arguments(self) -> List[str]:
        # Returns command line arguments that can be parsed via add_arguments_to_parser
        # and from_parsed_args to re-create the storage spec in a remote process
        ...

    @abc.abstractmethod
    def get_environment_variables(self) -> Iterable[kubernetes_client.V1EnvVar]:
        # Environment variables returned here will be added to the pod spec
        ...

    @classmethod
    @abc.abstractmethod
    def get_storage_type(cls) -> str:
        # This is used together with get_command_line_arguments so that a remote process
        # knows what sub-class of StorageBucketSpec to use to re-create the storage spec
        ...

    @classmethod
    @abc.abstractmethod
    def add_arguments_to_parser(cls, parser: argparse.ArgumentParser) -> None:
        # See get_command_line_arguments
        ...

    @classmethod
    @abc.abstractmethod
    async def from_parsed_args(cls, args: argparse.Namespace) -> AbstractStorageBucket:
        # See get_command_line_arguments
        ...


@dataclasses.dataclass(frozen=True)
class Kubernetes(Host):
    """
    Specifies a Kubernetes cluster to run a Meadowrun job on. resources_required is
    optional with the Kubernetes Host.

    Attributes:
        storage_spec: Specifies the object storage system to use. See derived classes of
            [StorageBucketSpec][meadowrun.StorageBucketSpec]. This can only be omitted
            if you are only using run_command and you've specified a specific container
            image to run on (i.e. rather than an EnvironmentSpec of some sort)
        kube_config_context: Specifies the kube config context to use. Default is None
            which means use the current context (i.e. `kubectl config current-context`)
        kubernetes_namespace: The Kubernetes namespace that Meadowrun will create Jobs
            in. This should usually not be left to the default value ("default") for any
            "real" workloads.
        resuable_pods: When set to True, starts generic long-lived pods that can be
            reused for multiple jobs. When set to False, starts a new pod(s) for every
            job
        pod_customization: A function like pod_customization(pod_template_spec) that
            will be called on the PodTemplateSpec just before we submit it to
            Kubernetes. You can make changes like specifying a serviceAccountName,
            adding ephemeral storage, etc. You can either modify the pod_template_spec
            argument in place and return it as the result, or construct a new
            [V1PodTemplateSpec](https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1PodTemplateSpec.md)
            and return that.
    """

    storage_spec: Optional[StorageBucketSpec]
    kube_config_context: Optional[str] = None
    kubernetes_namespace: str = "default"
    reusable_pods: bool = True
    pod_customization: Optional[
        Callable[
            [kubernetes_client.V1PodTemplateSpec], kubernetes_client.V1PodTemplateSpec
        ]
    ] = None

    async def get_storage_bucket(self) -> AbstractStorageBucket:
        # get the storage client

        # this is kind of weird, this should be called before any Kubernetes function,
        # but for now, _get_storage_bucket is always the first thing that's called
        await kubernetes_config.load_kube_config(context=self.kube_config_context)

        if self.storage_spec is None:
            raise ValueError(
                "The functionality you are trying to use requires specifying a "
                "storage_bucket. Please specify a storage_spec."
            )

        return await self.storage_spec.get_storage_bucket(self.kubernetes_namespace)

    async def run_job(
        self,
        resources_required: Optional[ResourcesInternal],
        job: Job,
        wait_for_result: WaitOption,
    ) -> JobCompletion[Any]:
        # TODO add support for this feature
        if job.sidecar_containers:
            raise NotImplementedError(
                "Sidecar containers are not yet supported for Kubernetes"
            )
        (
            is_custom_container_image,
            image_name,
            image_pull_secret_name,
            modified_job,
        ) = await _image_name_from_job(job)

        # now, if we have a custom container, try to run the job as a direct command. If
        # we don't have a custom container we "can't" run a direct command. We
        # theoretically could with a generic container if we asked for a preinstalled
        # interpreter, but there doesn't seem to be any real point in doing that.
        if is_custom_container_image:
            result = await self._run_direct_command_if_possible(
                job,
                resources_required,
                1,
                image_name,
                image_pull_secret_name,
                wait_for_result,
            )
            if result is not None:
                return result

        # if _run_direct_command_if_possible returns None, our job can't be run as a
        # direct command, so we need to run via run_job_local_storage_main.py
        return await self._run_regular_job(
            modified_job,
            image_name,
            image_pull_secret_name,
            resources_required,
            wait_for_result,
        )

    async def _run_direct_command_if_possible(
        self,
        job: Job,
        resources_required: Optional[ResourcesInternal],
        indexed_completions: Optional[int],
        image_name: str,
        image_pull_secret_name: Optional[str],
        wait_for_result: WaitOption,
    ) -> Optional[JobCompletion[None]]:
        # The normal way to run something via Meadowrun on Kubernetes is to call
        # run_local_storage_main in a pod, which will then start another process. For
        # very simple command jobs, we take a simpler approach of running a command
        # directly in a container rather than using run_local_storage_main. The main
        # advantage of this approach is that it doesn't require that Meadowrun is
        # installed in the container.

        if job.WhichOneof("job_spec") != "py_command":
            return None  # this obviously only works with commands

        if job.py_command.pickled_context_variables:
            return None  # pickled_context_variables requires Meadowrun is installed

        if indexed_completions is not None and indexed_completions != 1:
            # we should never get a py_command with indexed_completions, but just in
            # case
            return None

        if (
            job.WhichOneof("code_deployment") != "server_available_folder"
            or len(job.server_available_folder.code_paths) > 0
        ):
            # We don't support running a direct command with any sort of code
            # deployment. We could support a little more by adding
            # server_available_folder.code_paths to PATH and PYTHONPATH, but it's not
            # completely trivial to do that without interfering with whatever
            # environment variables have been set up in the container
            return None

        # run the job
        if not job.py_command.command_line:
            raise ValueError("command_line must have at least one string")
        command = list(job.py_command.command_line)

        environment_variables = {"PYTHONUNBUFFERED": "1"}
        environment_variables.update(**_string_pairs_to_dict(job.environment_variables))

        try:
            async with kubernetes_client.ApiClient() as api_client:
                return_codes = await _run_kubernetes_job(
                    kubernetes_client.CoreV1Api(api_client),
                    kubernetes_client.BatchV1Api(api_client),
                    job.job_id,
                    self.kubernetes_namespace,
                    image_name,
                    command,
                    environment_variables,
                    self.storage_spec,
                    image_pull_secret_name,
                    1,
                    [int(p) for p in expand_ports(job.ports)],
                    resources_required,
                    wait_for_result,
                )
        except asyncio.CancelledError:
            raise
        except Exception as e:
            raise MeadowrunException(
                ProcessState(
                    state=ProcessState.ProcessStateEnum.RUN_REQUEST_FAILED,
                )
            ) from e

        if len(return_codes) != 1:
            raise ValueError(
                "Programming error: requested 1 job but got back "
                f"{len(return_codes)}"
            )

        if return_codes[0] != 0:
            raise MeadowrunException(
                ProcessState(
                    state=ProcessState.ProcessStateEnum.NON_ZERO_RETURN_CODE,
                    return_code=return_codes[0],
                )
            )

        # TODO get the name of the pod that ran the job? It won't be super useful
        # because we delete the pod right away
        return JobCompletion(
            None,
            ProcessState.ProcessStateEnum.SUCCEEDED,
            "",
            0,
            "kubernetes",
        )

    async def _run_regular_job(
        self,
        job: Job,
        image_name: str,
        image_pull_secret_name: Optional[str],
        resources_required: Optional[ResourcesInternal],
        wait_for_result: WaitOption,
    ) -> JobCompletion[Any]:
        # This is the "normal" way to run jobs on Kubernetes. We assume that the image
        # specified by image_name has Meadowrun set up in it. We use the storage_bucket
        # to send a Job object via S3-compatible object storage and run
        # run_job_local_storage_main which will actually run our job. The results will
        # also come back via the object storage.

        # keeps track of most of what we write to object storage so we can clean it up
        # when we're done
        storage_keys_used = []

        if self.reusable_pods:
            process: KubernetesRemoteProcesses = ReusablePodRemoteProcesses()
        else:
            process = SingleUsePodRemoteProcesses()

        async with await self.get_storage_bucket() as storage_bucket, kubernetes_client.ApiClient() as api_client, kubernetes_stream.WsApiClient() as ws_api_client:  # noqa: E501
            # just for mypy, get_storage_bucket requires this
            assert self.storage_spec is not None

            core_api = kubernetes_client.CoreV1Api(api_client)
            ws_core_api = kubernetes_client.CoreV1Api(ws_api_client)
            batch_api = kubernetes_client.BatchV1Api(api_client)

            try:
                job_to_run_key = storage_key_job_to_run(job.job_id)
                storage_keys_used.append(job_to_run_key)
                await storage_bucket.write_bytes(
                    job.SerializeToString(), job_to_run_key
                )

                try:
                    return_codes = await (
                        await process.run(
                            core_api,
                            batch_api,
                            ws_core_api,
                            self.kubernetes_namespace,
                            image_name,
                            image_pull_secret_name,
                            [int(p) for p in expand_ports(job.ports)],
                            resources_required,
                            self.storage_spec,
                            job.job_id,
                            1,
                            wait_for_result,
                            self.pod_customization,
                        )
                    )
                except asyncio.CancelledError:
                    raise
                except BaseException as e:
                    raise MeadowrunException(
                        ProcessState(
                            state=ProcessState.ProcessStateEnum.RUN_REQUEST_FAILED,
                        )
                    ) from e

                if return_codes is not None:
                    if len(return_codes) != 1:
                        raise ValueError(
                            "Programming error, expected 1 return_codes but got "
                            f"{len(return_codes)}"
                        )
                    return_code = return_codes[0]
                    if return_code != 0:
                        raise MeadowrunException(
                            ProcessState(
                                state=(
                                    ProcessState.ProcessStateEnum.NON_ZERO_RETURN_CODE
                                ),
                                return_code=return_code,
                            )
                        )
                    # this should be 0 in an ideal world, but seems worth it to have one
                    # retry
                    timeout_seconds = 1
                else:
                    # TODO make this configurable
                    timeout_seconds = 60 * 60 * 24

                job_spec_type = job.WhichOneof("job_spec")
                if job_spec_type is None:
                    raise ValueError("Unexpected, job.job_spec is None")
                result = await get_job_completion_from_process_state(
                    storage_bucket,
                    job.job_id,
                    "0",
                    job_spec_type,
                    timeout_seconds,
                    "kubernetes",  # TODO get the name of the pod that ran the job?
                )
                process.received_result(0)
                return result
            finally:
                await process.kill_all()
                # TODO we should separately periodically clean up these files in case we
                # aren't able to execute this finally block
                if storage_bucket is not None:
                    await asyncio.gather(
                        *(
                            storage_bucket.delete_object(storage_key)
                            for storage_key in storage_keys_used
                        ),
                        return_exceptions=True,
                    )

    async def run_map_as_completed(
        self,
        function: Callable[[_T], _U],
        args: Sequence[_T],
        resources_required_per_task: Optional[ResourcesInternal],
        job_fields: Dict[str, Any],
        num_concurrent_tasks: int,
        pickle_protocol: int,
        wait_for_result: WaitOption,
        max_num_task_attempts: int,
        retry_with_more_memory: bool,
    ) -> AsyncIterable[TaskResult[_U]]:
        # TODO add support for this feature
        if job_fields["sidecar_containers"]:
            raise NotImplementedError(
                "Sidecar containers are not yet supported for Kubernetes"
            )
        if max_num_task_attempts != 1:
            raise NotImplementedError("max_num_task_attempts must be 1 on Kubernetes")

        async with await self.get_storage_bucket() as storage_bucket:
            # pretty much copied from AllocVM.run_map_as_completed

            driver = KubernetesGridJobDriver(self, num_concurrent_tasks, storage_bucket)

            # this should be in get_results, but with indexed workers we need to make
            # sure the tasks are uploaded before we can start workers
            await driver._add_tasks(args)

            try:
                run_worker_loops = asyncio.create_task(
                    driver.run_worker_functions(
                        function,
                        len(args),
                        resources_required_per_task,
                        job_fields,
                        pickle_protocol,
                        wait_for_result,
                    )
                )

                num_tasks_done = 0
                async for result in driver.get_results(args):
                    yield result
                    num_tasks_done += 1

                await run_worker_loops

            finally:
                keys_to_delete = [
                    storage_key_task_args(driver._job_id),
                    storage_key_ranges(driver._job_id),
                ]
                await asyncio.gather(
                    *(storage_bucket.delete_object(key) for key in keys_to_delete),
                    return_exceptions=True,
                )

        # this is for extra safety--the only case where we don't get all of our results
        # back should be if run_worker_loops throws an exception because there were
        # worker failures
        if num_tasks_done < len(args):
            raise ValueError(
                "Gave up retrieving task results, most likely due to worker failures. "
                f"Received {num_tasks_done}/{len(args)} task results."
            )


class KubernetesGridJobDriver:
    """
    Similar to GridJobDriver, should potentially be merged with that code at some point
    """

    def __init__(
        self,
        kubernetes: Kubernetes,
        num_concurrent_tasks: int,
        storage_bucket: AbstractStorageBucket,
    ):
        self._kubernetes = kubernetes

        # properties of the job
        self._num_concurrent_tasks = num_concurrent_tasks
        self._storage_bucket = storage_bucket

        self._job_id = str(uuid.uuid4())
        print(f"GridJob id is {self._job_id}")

        # run_worker_functions will set this to indicate to get_results
        # that there all of our workers have either exited unexpectedly (and we have
        # given up trying to restore them), or have been told to shutdown normally
        self._no_workers_available = asyncio.Event()

        # these events aren't actually used right now, but for now we're keeping this
        # code similar to GridJobDriver with the goal of eventually merging these
        # classes
        self._workers_needed = num_concurrent_tasks
        self._workers_needed_changed = asyncio.Event()

        self._abort_launching_new_workers = asyncio.Event()

        self._worker_process_states: List[List[WorkerProcessState]] = []
        self._worker_process_state_received = asyncio.Event()

    # these three functions are effectively the GridJobCloudInterface

    async def _add_tasks(self, args: Sequence[Any]) -> None:
        ranges = await upload_task_args(self._storage_bucket, self._job_id, args)
        # this is a hack--"normally" this would get sent with the "task assignment"
        # message, but we don't have the infrastructure for that in the case of Indexed
        # Jobs (static task-to-worker assignment)
        await self._storage_bucket.write_bytes(
            pickle.dumps(ranges),
            storage_key_ranges(self._job_id),
        )

    async def _receive_task_results(
        self, *, stop_receiving: asyncio.Event, workers_done: asyncio.Event
    ) -> AsyncIterable[Tuple[List[TaskProcessState], List[WorkerProcessState]]]:
        return receive_results(
            self._storage_bucket,
            self._job_id,
            stop_receiving=stop_receiving,
            all_workers_exited=workers_done,
            initial_wait_seconds=2,
            read_worker_process_states=self._kubernetes.reusable_pods,
        )

    async def _retry_task(self, task_id: int, attempts_so_far: int) -> None:
        raise NotImplementedError("Retries are not implemented for Kubernetes")

    def _worker_function_job(
        self,
        function: Callable[[_T], _U],
        num_args: int,
        job_fields: Dict[str, Any],
        pickle_protocol: int,
    ) -> Job:

        indexed_map_worker_args = num_args, self._num_concurrent_tasks, self._job_id

        return Job(
            job_id=self._job_id,
            py_agent=PyAgentJob(
                pickled_function=cloudpickle.dumps(function, protocol=pickle_protocol),
                pickled_agent_function=cloudpickle.dumps(
                    _indexed_map_worker, protocol=pickle_protocol
                ),
                pickled_agent_function_arguments=cloudpickle.dumps(
                    (indexed_map_worker_args, {}), protocol=pickle_protocol
                ),
            ),
            **job_fields,
        )

    async def run_worker_functions(
        self,
        function: Callable[[_T], _U],
        num_args: int,
        resources_required_per_task: Optional[ResourcesInternal],
        job_fields: Dict[str, Any],
        pickle_protocol: int,
        wait_for_result: WaitOption,
    ) -> None:
        # TODO implement wait_for_result options

        if self._kubernetes.reusable_pods:
            all_processes: KubernetesRemoteProcesses = ReusablePodRemoteProcesses()
        else:
            all_processes = SingleUsePodRemoteProcesses()

        worker_launch_task = None
        worker_process_state_received_task = asyncio.create_task(
            self._worker_process_state_received.wait()
        )
        workers_needed_changed_task = asyncio.create_task(
            self._workers_needed_changed.wait()
        )

        async with kubernetes_client.ApiClient() as api_client, kubernetes_stream.WsApiClient() as ws_api_client:  # noqa: E501
            try:
                # just for mypy, get_storage_bucket requires this
                assert self._kubernetes.storage_spec is not None

                job = self._worker_function_job(
                    function, num_args, job_fields, pickle_protocol
                )
                (
                    is_custom_container_image,
                    image_name,
                    image_pull_secret_name,
                    modified_job,
                ) = await _image_name_from_job(job)

                await self._storage_bucket.write_bytes(
                    modified_job.SerializeToString(),
                    storage_key_job_to_run(self._job_id),
                )

                core_api = kubernetes_client.CoreV1Api(api_client)
                ws_core_api = kubernetes_client.CoreV1Api(ws_api_client)
                batch_api = kubernetes_client.BatchV1Api(api_client)

                worker_launch_task = await all_processes.run(
                    core_api,
                    batch_api,
                    ws_core_api,
                    self._kubernetes.kubernetes_namespace,
                    image_name,
                    image_pull_secret_name,
                    [int(p) for p in expand_ports(job_fields["ports"])],
                    resources_required_per_task,
                    self._kubernetes.storage_spec,
                    self._job_id,
                    self._num_concurrent_tasks,
                    wait_for_result,
                    self._kubernetes.pod_customization,
                )

                # wait for all of the tasks to complete. If a worker fails, raise an
                # exception
                while self._workers_needed > 0:
                    tasks_to_wait: List[Task] = [
                        worker_process_state_received_task,
                        workers_needed_changed_task,
                    ]
                    if worker_launch_task is not None:
                        tasks_to_wait.append(worker_launch_task)
                    await asyncio.wait(
                        tasks_to_wait,
                        return_when=asyncio.FIRST_COMPLETED,
                    )

                    if worker_process_state_received_task.done():
                        while self._worker_process_states:
                            for (
                                worker_process_state
                            ) in self._worker_process_states.pop():
                                all_processes.received_result(
                                    int(worker_process_state.worker_index)
                                )

                                # TODO periodically also check for our pods disappearing
                                if (
                                    worker_process_state.result.state
                                    != ProcessState.ProcessStateEnum.SUCCEEDED
                                ):
                                    # TODO we should try to replace workers rather than
                                    # just throwing an exception immediately
                                    raise MeadowrunException(
                                        worker_process_state.result
                                    )

                        self._worker_process_state_received.clear()
                        worker_process_state_received_task = asyncio.create_task(
                            self._worker_process_state_received.wait()
                        )

                    if workers_needed_changed_task.done():
                        self._workers_needed_changed.clear()
                        workers_needed_changed_task = asyncio.create_task(
                            self._workers_needed_changed.wait()
                        )

                    if worker_launch_task is not None and worker_launch_task.done():
                        # first, this will raise any exceptions from this task. Then, we
                        # check the result. If the result is not None, that means the
                        # workers launched and ran to completion and we got back a
                        # return code, so we should exit this loop. If we got back a
                        # None, that means the workers launched successfully (as far as
                        # we can tell), and we're not planning on getting notified when
                        # they finish. If we got rid of the single-use pods option, we
                        # could get rid of the break, but we would still want to call
                        # .result() here to make sure we see any exceptions
                        if worker_launch_task.result() is not None:
                            break

                        # don't keep including this in asyncio.wait
                        worker_launch_task = None
            finally:
                try:
                    self._no_workers_available.set()

                    # make sure we kill any remote processes and finish awaiting any
                    # deallocation tasks
                    await all_processes.kill_all()

                    worker_process_state_received_task.cancel()
                    workers_needed_changed_task.cancel()
                    if worker_launch_task is not None:
                        worker_launch_task.cancel()

                    await self._storage_bucket.delete_object(
                        storage_key_job_to_run(self._job_id)
                    )
                except asyncio.CancelledError:
                    raise
                except BaseException:
                    pass

    async def get_results(self, args: Sequence[_T]) -> AsyncIterable[TaskResult]:
        """Yields TaskResult objects as soon as tasks complete."""

        # semi-copy/paste from GridJobDriver.add_tasks_and_get_results

        num_tasks_done = 0
        # stop_receiving tells _cloud_interface.receive_task_results that there are no
        # more results to get
        stop_receiving = asyncio.Event()
        if len(args) == num_tasks_done:
            stop_receiving.set()
        last_printed_update = time.time()
        print(
            f"Waiting for task results. Requested: {len(args)}, "
            f"Done: {num_tasks_done}"
        )
        async for task_batch, worker_batch in await self._receive_task_results(
            stop_receiving=stop_receiving, workers_done=self._no_workers_available
        ):
            for task_process_state in task_batch:
                task_result = TaskResult.from_process_state(task_process_state)
                if task_result.is_success:
                    num_tasks_done += 1
                    yield task_result
                else:
                    print(f"Task {task_process_state.task_id} failed")
                    num_tasks_done += 1
                    yield task_result

            if worker_batch:
                self._worker_process_states.append(worker_batch)
                self._worker_process_state_received.set()

            if num_tasks_done >= len(args):
                stop_receiving.set()
            else:
                t0 = time.time()
                if t0 - last_printed_update > _PRINT_RECEIVED_TASKS_SECONDS:
                    print(
                        f"Waiting for task results. Requested: {len(args)}, "
                        f"Done: {num_tasks_done}"
                    )
                    last_printed_update = t0

            # reduce the number of workers needed if we have more workers than
            # outstanding tasks
            num_workers_needed = max(len(args) - num_tasks_done, 0)
            if num_workers_needed < self._workers_needed:
                self._workers_needed = num_workers_needed
                self._workers_needed_changed.set()

        if num_tasks_done < len(args):
            # It would make sense for this to raise an exception, but it's more helpful
            # to see the actual worker failures, and run_worker_functions should always
            # raise an exception in that case. The caller should still check though that
            # we returned all of the task results we were expecting.
            print(
                "Gave up retrieving task results, most likely due to worker failures. "
                f"Received {num_tasks_done}/{len(args)} task results."
            )
        else:
            print(f"Received all {len(args)} task results.")


def _resources_to_kubernetes(resources: ResourcesInternal) -> Dict[str, int]:
    result = {}

    if LOGICAL_CPU in resources.consumable:
        result["cpu"] = math.ceil(resources.consumable[LOGICAL_CPU])
    if MEMORY_GB in resources.consumable:
        result["memory"] = math.ceil(resources.consumable[MEMORY_GB] * (1024**3))
    if GPU in resources.consumable:
        num_gpus = math.ceil(resources.consumable[GPU])
        if "nvidia" in resources.non_consumable:
            result["nvidia.com/gpu"] = num_gpus
        else:
            raise ValueError(
                "Must specify a type of GPU (e.g. nvidia) if a GPU resource is "
                "requested"
            )

    # TODO maybe warn if people are trying to use resources that we don't know how to
    # interpret
    # TODO maybe turn max_eviction_rate into a pod disruption budget?

    return result


def _get_additional_container_parameters(
    ports: List[int], resources: Optional[ResourcesInternal]
) -> Dict[str, Any]:
    additional_container_parameters = {}
    if ports:
        additional_container_parameters["ports"] = [
            kubernetes_client.V1ContainerPort(container_port=port) for port in ports
        ]
    if resources is not None:
        additional_container_parameters[
            "resources"
        ] = kubernetes_client.V1ResourceRequirements(
            requests=_resources_to_kubernetes(resources)
        )
    return additional_container_parameters


def _pod_meets_requirements(
    pod: kubernetes_client.V1Pod,
    ports: List[int],
    resources: Optional[ResourcesInternal],
) -> bool:
    main_containers = [
        container for container in pod.spec.containers if container.name == "main"
    ]
    if len(main_containers) != 1:
        print(
            f"Warning: pod {pod.metadata.name} has {len(main_containers)} containers "
            f"named main, which was unexpected"
        )
        return False
    main_container = main_containers[0]

    if ports:
        if not main_container.ports:
            return False
        required_ports = set(ports)
        for port in main_container.ports:
            required_ports.discard(port.container_port)
        if required_ports:
            return False

    if resources:
        existing_resources = main_container.resources.requests
        for key, value in _resources_to_kubernetes(resources).items():
            if key not in existing_resources:
                return False
            # this str conversion is a bit sketchy. Also, we could do a >= check here,
            # but it seems better to just get the exact same resource requirements--the
            # existing pod will just disappear on its own
            if existing_resources[key] != str(value):
                return False

    return True


def _add_meadowrun_variables_to_environment(
    environment: List[kubernetes_client.V1EnvVar],
    environment_variables: Dict[str, str],
    storage_spec: Optional[StorageBucketSpec],
) -> None:
    """
    Modifies environment in place!! environment_variables is just to make sure we don't
    overwrite an existing environment variable
    """

    if storage_spec is not None:
        for env_var in storage_spec.get_environment_variables():
            if env_var.name not in environment_variables:
                environment.append(env_var)

    if MEADOWRUN_POD_NAME not in environment_variables:
        environment.append(
            kubernetes_client.V1EnvVar(
                name=MEADOWRUN_POD_NAME,
                value_from=kubernetes_client.V1EnvVarSource(
                    field_ref=kubernetes_client.V1ObjectFieldSelector(
                        field_path="metadata.name"
                    )
                ),
            )
        )


class KubernetesRemoteProcesses(abc.ABC):
    """An interface for running a job on Kubernetes"""

    @abc.abstractmethod
    async def run(
        self,
        core_api: kubernetes_client.CoreV1Api,
        batch_api: kubernetes_client.BatchV1Api,
        ws_core_api: kubernetes_client.CoreV1Api,
        kubernetes_namespace: str,
        image_name: str,
        image_pull_secret_name: Optional[str],
        ports: List[int],
        resources: Optional[ResourcesInternal],
        storage_spec: StorageBucketSpec,
        job_id: str,
        # TODO this will probably eventually need to be replaced with an explicit list
        # of worker indexes if we want to be able to restart workers
        num_executions: int,
        wait_for_result: WaitOption,
        pod_customization: Optional[
            Callable[
                [kubernetes_client.V1PodTemplateSpec],
                kubernetes_client.V1PodTemplateSpec,
            ]
        ] = None,
    ) -> Task[Optional[Sequence[int]]]:
        """
        Assumes that the Job has already been uploaded to object storage.

        If this returns a Sequence[int], that means the jobs ran to completion and we
        are returning the exit codes. If this returns None, that means we just launched
        the jobs and then detached, and the caller is responsible for figuring out when
        the job completes.
        """
        ...

    @abc.abstractmethod
    def received_result(self, worker_index: int) -> None:
        """
        This notifies us that the specified worker completed. We will do any
        deallocation necessary and mark this as a worker that we don't need to
        explicitly kill.
        """
        ...

    @abc.abstractmethod
    def kill_all(self) -> Awaitable[Any]:
        """
        Kills all of the workers. This function MUST be called after run in a finally
        block, even if we've successfully received results for all of the workers.

        Return type should be None but it's more trouble than it's worth to make that
        work in the implementation
        """
        ...


class SingleUsePodRemoteProcesses(KubernetesRemoteProcesses):
    def __init__(self) -> None:
        self.run_task: Optional[Task[Optional[Sequence[int]]]] = None
        self.job_has_finished: bool = False

    async def run(
        self,
        core_api: kubernetes_client.CoreV1Api,
        batch_api: kubernetes_client.BatchV1Api,
        ws_core_api: kubernetes_client.CoreV1Api,
        kubernetes_namespace: str,
        image_name: str,
        image_pull_secret_name: Optional[str],
        ports: List[int],
        resources: Optional[ResourcesInternal],
        storage_spec: StorageBucketSpec,
        job_id: str,
        num_executions: int,
        wait_for_result: WaitOption,
        pod_customization: Optional[
            Callable[
                [kubernetes_client.V1PodTemplateSpec],
                kubernetes_client.V1PodTemplateSpec,
            ]
        ] = None,
    ) -> Task[Optional[Sequence[int]]]:
        command = [
            "python",
            "-m",
            "meadowrun.run_job_local_storage_main",
            "--job-id",
            job_id,
        ] + storage_spec.get_command_line_arguments()

        self.run_task = asyncio.create_task(
            _run_kubernetes_job(
                core_api,
                batch_api,
                job_id,
                kubernetes_namespace,
                image_name,
                command,
                {"PYTHONUNBUFFERED": "1"},
                storage_spec,
                image_pull_secret_name,
                num_executions,
                ports,
                resources,
                wait_for_result,
                pod_customization,
            )
        )

        return self.run_task

    def received_result(self, worker_index: int) -> None:
        # this is fine for now, but when we implement kill_all correctly, we will want
        # to keep track of these completions
        self.job_has_finished = True
        if self.run_task is not None:
            self.run_task.cancel()

    async def kill_all(self) -> None:
        if not self.job_has_finished:
            if self.run_task is not None:
                # this works right now because _run_kubernetes_job has a finally block
                # that always deletes the job it creates
                self.run_task.cancel()


async def _run_kubernetes_job(
    core_api: kubernetes_client.CoreV1Api,
    batch_api: kubernetes_client.BatchV1Api,
    job_id: str,
    kubernetes_namespace: str,
    image: str,
    args: List[str],
    environment_variables: Dict[str, str],
    storage_spec: Optional[StorageBucketSpec],
    image_pull_secret_name: Optional[str],
    indexed_completions: Optional[int],
    ports: List[int],
    resources: Optional[ResourcesInternal],
    wait_for_result: WaitOption,
    pod_customization: Optional[
        Callable[
            [kubernetes_client.V1PodTemplateSpec],
            kubernetes_client.V1PodTemplateSpec,
        ]
    ] = None,
) -> Sequence[int]:
    """
    Runs the specified job on Kubernetes, waits for it to complete, and returns the exit
    code
    """

    # TODO add support for DO_NOT_WAIT
    if wait_for_result == WaitOption.DO_NOT_WAIT:
        raise NotImplementedError(
            f"{wait_for_result} is not supported for Kubernetes yet"
        )

    # create the job

    environment = [
        kubernetes_client.V1EnvVar(name=key, value=value)
        for key, value in environment_variables.items()
    ]

    _add_meadowrun_variables_to_environment(
        environment, environment_variables, storage_spec
    )

    if indexed_completions:
        additional_job_spec_parameters = {
            "completions": indexed_completions,
            "parallelism": indexed_completions,
            "completion_mode": "Indexed",
        }
    else:
        additional_job_spec_parameters = {}

    if image_pull_secret_name:
        additional_pod_spec_parameters = {
            "image_pull_secrets": [
                kubernetes_client.V1LocalObjectReference(image_pull_secret_name)
            ]
        }
    else:
        additional_pod_spec_parameters = {}

    # https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Job.md
    pod_template_spec = kubernetes_client.V1PodTemplateSpec(
        spec=kubernetes_client.V1PodSpec(
            containers=[
                kubernetes_client.V1Container(
                    name="main",
                    image=image,
                    args=args,
                    env=environment,
                    **_get_additional_container_parameters(ports, resources),
                )
            ],
            restart_policy="Never",
            **additional_pod_spec_parameters,
        )
    )
    if pod_customization:
        pod_template_spec = pod_customization(pod_template_spec)
    body = kubernetes_client.V1Job(
        metadata=kubernetes_client.V1ObjectMeta(name=job_id),
        spec=kubernetes_client.V1JobSpec(
            # we also try to delete manually. This field could be ignored if the TTL
            # controller is not available
            ttl_seconds_after_finished=10,
            backoff_limit=0,
            template=pod_template_spec,
            **additional_job_spec_parameters,
        ),
    )

    try:
        # this returns a result, but we don't really need anything from it
        await batch_api.create_namespaced_job(kubernetes_namespace, body)

        # Now that we've created the job, wait for the pods to be created

        if indexed_completions:
            pod_generate_names = [f"{job_id}-{i}-" for i in range(indexed_completions)]
        else:
            pod_generate_names = [f"{job_id}-"]

        pods = await get_pods_for_job(
            core_api, kubernetes_namespace, job_id, pod_generate_names
        )

        pod_names = [pod.metadata.name for pod in pods]
        print(f"Created pod(s) {', '.join(pod_names)} for job {job_id}")

        # Now that the pods have been created, stream their logs and get their exit
        # codes

        return await asyncio.gather(
            *[
                asyncio.create_task(
                    wait_for_pod(
                        core_api, job_id, kubernetes_namespace, pod, wait_for_result
                    )
                )
                for pod in pods
            ]
        )
    finally:
        try:
            await batch_api.delete_namespaced_job(
                job_id, kubernetes_namespace, propagation_policy="Foreground"
            )
        except kubernetes_client_exceptions.ApiException as e:
            print(f"Warning, error cleaning up job: {e}")


class ReusablePodRemoteProcesses(KubernetesRemoteProcesses):
    def __init__(self) -> None:
        self.all_processes: List[ReusablePodRemoteProcess] = []
        self.deallocation_tasks: List[Task[None]] = []

    async def run(
        self,
        core_api: kubernetes_client.CoreV1Api,
        batch_api: kubernetes_client.BatchV1Api,
        ws_core_api: kubernetes_client.CoreV1Api,
        kubernetes_namespace: str,
        image_name: str,
        image_pull_secret_name: Optional[str],
        ports: List[int],
        resources: Optional[ResourcesInternal],
        storage_spec: StorageBucketSpec,
        job_id: str,
        num_executions: int,
        wait_for_result: WaitOption,
        pod_customization: Optional[
            Callable[
                [kubernetes_client.V1PodTemplateSpec],
                kubernetes_client.V1PodTemplateSpec,
            ]
        ] = None,
    ) -> Task[Optional[Sequence[int]]]:

        storage_spec_args = (
            f"{STORAGE_TYPE} {storage_spec.get_storage_type()} "
            + " ".join(
                shlex.quote(arg) for arg in storage_spec.get_command_line_arguments()
            )
        )
        # TODO it's possible that a job will run more than once on the same container,
        # although not in parallel, in that case the job log will be overwritten
        _COMMAND_TEMPLATE = (
            "PYTHONUNBUFFERED=1 MEADOWRUN_WORKER_INDEX={worker_index} python -m "
            f"meadowrun.run_job_local_storage_main --job-id {job_id} "
            f"{storage_spec_args} >/var/meadowrun/job_logs/{job_id}.log 2>&1 & echo $$"
        )
        # echo $$ gets the process id of the current bash session (we will run
        # inner_command in a bash session). We assume that all processes started in this
        # session will get this id as their process group id (it's not clear if this
        # assumption is bulletproof) so that we can kill them all quickly. (100%
        # reliable would be to do echo $! to get the run_job_local_storage_main PID and
        # then explicitly ask for its group id and kill all processes in that group)

        process_group_id_tasks = []

        worker_index = 0
        async for pods in _get_meadowrun_reusable_pods(
            kubernetes_namespace,
            image_name,
            image_pull_secret_name,
            ports,
            resources,
            num_executions,
            core_api,
            batch_api,
            storage_spec,
            pod_customization,
        ):

            for pod in pods:
                command = [
                    "/bin/bash",
                    "-c",
                    _COMMAND_TEMPLATE.format(worker_index=worker_index),
                ]

                process_group_id_task = asyncio.create_task(
                    run_command_on_pod(
                        pod.metadata.name, kubernetes_namespace, command, ws_core_api
                    )
                )

                process_group_id_tasks.append(process_group_id_task)
                self.all_processes.append(
                    ReusablePodRemoteProcess(
                        core_api,
                        ws_core_api,
                        kubernetes_namespace,
                        pod,
                        process_group_id_task,
                    )
                )
                worker_index += 1

        return asyncio.create_task(_gather_return_none(process_group_id_tasks))

    def received_result(self, worker_index: int) -> None:
        self.deallocation_tasks.append(
            asyncio.create_task((self.all_processes[worker_index]).received_result())
        )

    def kill_all(self) -> Awaitable:
        return asyncio.gather(
            *(process.kill() for process in self.all_processes),
            # executions where we've gotten the result already don't need to be killed
            # (because the fact that we received a result means they're presumably dead)
            # but we need to make sure their deallocation tasks are awaited
            *self.deallocation_tasks,
            return_exceptions=True,
        )


class ReusablePodRemoteProcess:
    def __init__(
        self,
        core_api: kubernetes_client.CoreV1Api,
        ws_core_api: kubernetes_client.CoreV1Api,
        kubernetes_namespace: str,
        pod: kubernetes_client.V1Pod,
        process_group_id_task: Awaitable[str],
    ):
        self._pod = pod
        self._kubernetes_namespace = kubernetes_namespace
        self._process_group_id_task = process_group_id_task
        self._core_api = core_api
        self._ws_core_api = ws_core_api

        self._has_exited = False

    def _deallocate(self) -> Coroutine[Any, Any, None]:
        # now mark those pods as ready--this is an optimization--the automated readiness
        # probe will mark the pods as ready eventually, but this is faster
        return set_main_container_ready(self._core_api, self._pod, True)

    async def received_result(self) -> None:
        # we got a result through some other means
        self._has_exited = True
        await self._deallocate()

    async def kill(self) -> None:
        if not self._has_exited:
            self._has_exited = True
            await run_command_on_pod(
                self._pod.metadata.name,
                self._kubernetes_namespace,
                ["kill", "-9", "--", f"-{await self._process_group_id_task}"],
                self._ws_core_api,
            )
            await self._deallocate()


async def _get_meadowrun_reusable_pods(
    kubernetes_namespace: str,
    image_name: str,
    image_pull_secret_name: Optional[str],
    ports: List[int],
    resources: Optional[ResourcesInternal],
    number_of_pods: int,
    core_api: kubernetes_client.CoreV1Api,
    batch_api: kubernetes_client.BatchV1Api,
    storage_spec: Optional[StorageBucketSpec],
    pod_customization: Optional[
        Callable[
            [kubernetes_client.V1PodTemplateSpec],
            kubernetes_client.V1PodTemplateSpec,
        ]
    ],
) -> AsyncIterable[List[kubernetes_client.V1Pod]]:

    # some potential confusion between different images, would it be better to hash
    # instead?
    image_name_label = image_name.replace("/", ".").replace(":", ".")

    pods_response = await core_api.list_namespaced_pod(
        kubernetes_namespace,
        # it would be nice to put ports and resources in the labels, but we need some
        # sort of string serialization format for that. For now we just filter them out
        # on the client side
        label_selector=f"meadowrun.io/image-name={image_name_label}",
        # Available fields:
        # https://github.com/kubernetes/kubernetes/blob/f01c9e8683adacbfbad58e5153dfac9ebf954c4b/pkg/registry/core/pod/strategy.go#L301
        field_selector="status.phase=Running",
    )
    additional_container_parameters = _get_additional_container_parameters(
        ports, resources
    )
    existing_pods: List[kubernetes_client.V1Pod] = []
    for pod in pods_response.items:
        if len(existing_pods) >= number_of_pods:
            break
        if (
            get_main_container_is_ready(pod)
            and _pod_meets_requirements(pod, ports, resources)
            and pod.metadata.deletion_timestamp is None
        ):
            # This is an "optimistic concurrency"-style check. It's possible that the
            # automated readiness probe will run between now and when we're able to
            # launch our job, remarking this pod as ready and allowing another job to
            # swoop in and steal the pod. In that case, the job will fail to start up
            # because it won't be able to get the _JOB_IS_RUNNING file lock, and we can
            # request another pod
            await set_main_container_ready(core_api, pod, False)
            existing_pods.append(pod)

    # limit may be ignored according to the Kubernetes API spec
    # TODO pods that say "terminating" in the dashboard show up as "running" in this
    # query
    if len(existing_pods) > number_of_pods:
        existing_pods = existing_pods[:number_of_pods]
    if existing_pods:
        print(
            f"Reusing {len(existing_pods)} existing pods: "
            + ", ".join(pod.metadata.name for pod in existing_pods)
        )
        yield existing_pods

    remaining_pods_to_launch = number_of_pods - len(existing_pods)

    if remaining_pods_to_launch > 0:
        environment: List[kubernetes_client.V1EnvVar] = []

        _add_meadowrun_variables_to_environment(environment, {}, storage_spec)

        if image_pull_secret_name:
            additional_pod_spec_parameters = {
                "image_pull_secrets": [
                    kubernetes_client.V1LocalObjectReference(image_pull_secret_name)
                ]
            }
        else:
            additional_pod_spec_parameters = {}

        job_name = f"mdr-reusable-{uuid.uuid4()}"
        # it would be way more convenient to do -m
        # meadowrun.k8s_integration.is_job_running, but that is way slower because it
        # imports everything (i.e. boto3)
        is_job_running_command = (
            "import runpy,site;runpy.run_path(site.getsitepackages()[0]"
            '+"/meadowrun/k8s_integration/is_job_running.py",run_name="__main__")'
        )
        pod_template_spec = kubernetes_client.V1PodTemplateSpec(
            metadata=kubernetes_client.V1ObjectMeta(
                labels={"meadowrun.io/image-name": image_name_label}
            ),
            spec=kubernetes_client.V1PodSpec(
                containers=[
                    kubernetes_client.V1Container(
                        name="main",
                        image=image_name,
                        args=[
                            "python",
                            "-m",
                            "meadowrun.k8s_integration.k8s_main",
                        ],
                        env=environment,
                        readiness_probe=kubernetes_client.V1Probe(
                            _exec=kubernetes_client.V1ExecAction(
                                command=["python", "-c", is_job_running_command]
                            ),
                            initial_delay_seconds=15,
                            period_seconds=15,
                            timeout_seconds=5,
                            failure_threshold=1,
                            success_threshold=1,
                        ),
                        **additional_container_parameters,
                    )
                ],
                restart_policy="Never",
                **additional_pod_spec_parameters,
            ),
        )
        if pod_customization:
            pod_template_spec = pod_customization(pod_template_spec)
        body = kubernetes_client.V1Job(
            metadata=kubernetes_client.V1ObjectMeta(name=job_name),
            spec=kubernetes_client.V1JobSpec(
                ttl_seconds_after_finished=10,
                backoff_limit=0,
                template=pod_template_spec,
                completion_mode="Indexed",
                completions=remaining_pods_to_launch,
                parallelism=remaining_pods_to_launch,
            ),
        )

        await batch_api.create_namespaced_job(kubernetes_namespace, body)

        try:
            pods = await get_pods_for_job(
                core_api,
                kubernetes_namespace,
                job_name,
                [f"{job_name}-{i}-" for i in range(remaining_pods_to_launch)],
            )
        except BaseException:
            # usually the ttl_seconds_after_finished will take care of deleting the job,
            # but there's a case where the pods can't be created (e.g. if you provide an
            # invalid serviceAccountName), and the job will stick around forever
            await batch_api.delete_namespaced_job(job_name, kubernetes_namespace)
            raise

        for pod_future in asyncio.as_completed(
            [
                wait_for_pod_running(core_api, job_name, kubernetes_namespace, pod)
                for pod in pods
            ]
        ):
            yield [await pod_future]

        print(f"Started {remaining_pods_to_launch} new pods")


async def _gather_return_none(tasks: List[Task]) -> None:
    """A workaround because you can't do asyncio.create_task(asyncio.gather(*tasks))"""
    # consider throwing all exceptions not just the first one we get
    await asyncio.gather(*tasks, return_exceptions=False)


STORAGE_TYPE = "--storage-type"



"""
Like __meadowrun_func_worker, but connections to given agent.
Listens for arguments, executes the function and returns the result.
"""

import asyncio

import importlib  # available in python 3.1+
import argparse  # available in python 3.2+
import pickle
import struct
import traceback
from typing import Callable


async def send_message(
    writer: asyncio.StreamWriter, state: str, result_bytes: bytes, pickle_protocol: int
) -> None:
    # it's important here that the result is pickled separately. If not, the agent's
    # receive_message will unpickle result. This is not only unnecessary, but also
    # potentially doesn't work because the agent doesn't run in the worker's
    # environment.
    msg_bs = pickle.dumps((state, result_bytes), protocol=pickle_protocol)
    msg_len = struct.pack(">i", len(msg_bs))
    writer.write(msg_len)
    writer.write(msg_bs)
    await writer.drain()


async def receive_bytes(reader: asyncio.StreamReader, bytes_len: int) -> bytearray:
    result_bs = bytearray()
    while len(result_bs) < bytes_len:
        result_bs.extend(await reader.read(bytes_len - len(result_bs)))
    return result_bs


async def do_tasks(
    function: Callable,
    pickle_protocol: int,
    reader: asyncio.StreamReader,
    writer: asyncio.StreamWriter,
) -> None:
    try:
        while True:
            arg_size_bs = await reader.read(4)
            if len(arg_size_bs) == 0:
                break
            (arg_size,) = struct.unpack(">i", arg_size_bs)

            try:
                if arg_size > 0:
                    arg_bs = await receive_bytes(reader, arg_size)
                    function_args, function_kwargs = pickle.loads(arg_bs)
                else:
                    function_args, function_kwargs = (), {}
                # run the function
                result = function(*(function_args or ()), **(function_kwargs or {}))
                result_bytes = pickle.dumps(result, protocol=pickle_protocol)
            except Exception as e:
                # first print the exception for the local log file
                traceback.print_exc()

                # next, send the exception back
                tb = "".join(traceback.format_exception(type(e), e, e.__traceback__))
                await send_message(
                    writer,
                    "PYTHON_EXCEPTION",
                    pickle.dumps((str(type(e)), str(e), tb), protocol=pickle_protocol),
                    pickle_protocol,
                )
            else:
                # send back results
                await send_message(writer, "SUCCEEDED", result_bytes, pickle_protocol)
    finally:
        # close connection
        writer.close()


async def connect_and_do_tasks(
    host: str, port: int, function: Callable, pickle_protocol: int
) -> None:
    reader, writer = await asyncio.open_connection(host, port)
    await do_tasks(function, pickle_protocol, reader, writer)


def get_function(args: argparse.Namespace) -> Callable:
    # import the module/unpickle the function
    if args.module_name is not None:
        print(f"About to import {args.module_name}.{args.function_name}")
        module = importlib.import_module(args.module_name)
        function = getattr(module, args.function_name)
        print(
            f"Imported {args.function_name} from "
            f"{getattr(module, '__file__', str(module))}"
        )
    else:
        pickled_function_path = args.io_path + ".function"
        print(f"Unpickling function from {pickled_function_path}")
        with open(pickled_function_path, "rb") as f:
            function = pickle.load(f)
    return function


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--module-name")
    parser.add_argument("--function-name")
    parser.add_argument("--io-path", required=True)
    parser.add_argument("--has-pickled-function", action="store_true")
    parser.add_argument("--result-highest-pickle-protocol", type=int, required=True)
    parser.add_argument("--host", required=True)
    parser.add_argument("--port", type=int, required=True)

    args = parser.parse_args()

    if bool(args.module_name) ^ bool(args.function_name):
        raise ValueError(
            "Cannot specify just one of --module-name and --function-name without the "
            "other"
        )
    if not (bool(args.module_name) ^ args.has_pickled_function):
        raise ValueError(
            "Must specify either --module-name with --function-name OR "
            "--has-pickled-function but not both"
        )

    # prepare filenames and pickle protocol for the result of the agent
    state_filename = args.io_path + ".state"
    result_filename = args.io_path + ".result"

    result_pickle_protocol = min(
        args.result_highest_pickle_protocol, pickle.HIGHEST_PROTOCOL
    )
    try:
        function = get_function(args)
        asyncio.run(
            connect_and_do_tasks(
                args.host,
                args.port,
                function,
                result_pickle_protocol,
            )
        )

    except Exception as e:
        # first print the exception for the local log file
        traceback.print_exc()

        # next, send the exception back
        tb = "".join(traceback.format_exception(type(e), e, e.__traceback__))
        # see MeadowRunClientAsync for why we don't just pickle the exception
        with open(state_filename, "w", encoding="utf-8") as state_text_writer:
            state_text_writer.write("PYTHON_EXCEPTION")
        with open(result_filename, "wb") as f:
            pickle.dump((str(type(e)), str(e), tb), f, protocol=result_pickle_protocol)
    else:
        # send back results
        with open(state_filename, "w", encoding="utf-8") as state_text_writer:
            state_text_writer.write("SUCCEEDED")
        with open(result_filename, "wb") as f:
            pickle.dump("agent exited normally", f, protocol=result_pickle_protocol)


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import os
import sys
from typing import Callable, Awaitable, Optional

import filelock

from meadowrun.shared import remove_corrupted_environment
from meadowrun.deployment.prerequisites import (
    EnvironmentSpecPrerequisites,
    GOOGLE_AUTH_PACKAGE,
)
from meadowrun.storage_keys import STORAGE_ENV_CACHE_PREFIX


# _pip_freeze and friends are run in the user environment and have to deal with
# arbitrary configurations. get_cached_or_create_pip_environment and
# create_pip_environment run in controlled environments, either in the Meadowrun docker
# image or a Meadowrun AMI


async def _pip_freeze(python_interpreter: str, *freeze_opts: str) -> str:
    """Effectively returns the output of:
    > <python_interpreter> -m pip freeze <freeze_opts>
    """

    # `pip freeze` and `pip list --format=freeze` usually give identical results, but
    # not for "packages installed via Direct URLs":
    # https://github.com/pypa/pip/issues/8176 which poetry uses to install all packages.
    # So for us, `pip list --format=freeze` gives us the output we want both with poetry
    # and in a "regular" virtualenv. We should keep an eye on that issue, though, as it
    # looks like this difference in behavior was not totally intentional and may change
    # in the future.
    # Update Aug 22: we try to find pyproject.toml and poetry.lock before getting here,
    # so we're using pip freeze now. Direct URLs are needed to support packages
    # installed via git URLs.
    env = os.environ.copy()
    env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
    opts = ("-m", "pip", "freeze") + freeze_opts
    p = await asyncio.create_subprocess_exec(
        python_interpreter,
        *opts,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        env=env,
    )
    stdout, stderr = await p.communicate()

    if p.returncode != 0:
        raise ValueError(
            f"pip freeze failed, return code {p.returncode}: {stderr.decode()}"
        )

    return stdout.decode()


async def pip_freeze_exclude_editable(
    python_interpreter: Optional[str] = None,
) -> str:
    """
    Returns the result of calling pip freeze on the specified python_interpreter, or the
    current one if not specified. Excludes editable installs - the local code deployment
    should take largely care of the editable installs, except for running setup.py or
    equivalent.
    """
    if python_interpreter is None:
        python_interpreter = sys.executable
    # TODO we have not actually implemented the portable part of this yet
    return await _pip_freeze(python_interpreter, "--exclude-editable")


_PIP_ENVIRONMENT_TIMEOUT = 10 * 60


async def get_cached_or_create_pip_environment(
    environment_hash: str,
    requirements_file_path: str,
    prerequisites: EnvironmentSpecPrerequisites,
    new_environment_path: str,
    try_get_file: Callable[[str, str], Awaitable[bool]],
    upload_file: Callable[[str, str], Awaitable[None]],
) -> str:
    """
    If the desired pip environment exists, does nothing. If the environment has been
    cached, creates it from the cache. Otherwise creates the environment from scratch
    and caches it. Returns the path to the newly created python interpreter.

    try_get_file and upload_file are for interacting with the cache, see
    compile_environment_spec_locally for more details
    """
    # this assumes that that current version of python is what will be used in
    # create_pip_environment
    new_environment_path = (
        f"{new_environment_path}-{sys.version_info.major}.{sys.version_info.minor}"
    )

    new_environment_interpreter = os.path.join(new_environment_path, "bin", "python")

    with filelock.FileLock(f"{new_environment_path}.lock", _PIP_ENVIRONMENT_TIMEOUT):
        if os.path.exists(new_environment_path):
            return new_environment_interpreter

        remote_cached_file_name = (
            f"{STORAGE_ENV_CACHE_PREFIX}{environment_hash}-{sys.version_info.major}."
            f"{sys.version_info.minor}.tar.gz"
        )
        local_cached_file = f"{new_environment_path}.tar.gz"
        download_succeeded = await try_get_file(
            remote_cached_file_name, local_cached_file
        )
        if download_succeeded:
            try:
                print("Unpacking cached pip environment")
                os.makedirs(new_environment_path, exist_ok=True)
                try:
                    # TODO maybe cleaner to use the built-in python tar libraries?
                    return_code = await (
                        await asyncio.create_subprocess_exec(
                            "tar", "-xzf", local_cached_file, "-C", new_environment_path
                        )
                    ).wait()
                    if return_code != 0:
                        raise ValueError(
                            f"Unpacking cached pip environment {local_cached_file} "
                            f"returned code {return_code}"
                        )
                    return new_environment_interpreter
                except BaseException:
                    remove_corrupted_environment(new_environment_path)
                    raise
            finally:
                try:
                    os.remove(local_cached_file)
                except asyncio.CancelledError:
                    raise
                except BaseException:
                    pass

        print("Creating the pip environment")
        try:
            await create_pip_environment(
                requirements_file_path, new_environment_path, prerequisites
            )
        except BaseException:
            remove_corrupted_environment(new_environment_path)
            raise

        try:
            import venv_pack  # see note on reference in pyproject.toml
        except ImportError:
            print(
                "Warning unable to cache pip environment because venv_pack is missing"
            )
            return new_environment_interpreter

        # TODO we shouldn't wait for this to start running the job but we also shouldn't
        # kill the container until this finishes
        print("Caching the pip environment")
        try:
            # there might be an old file from a failed attempt
            os.remove(local_cached_file)
        except OSError:
            pass
        venv_pack.pack(new_environment_path, output=local_cached_file)
        await upload_file(local_cached_file, remote_cached_file_name)

        return new_environment_interpreter


async def create_pip_environment(
    requirements_file_path: str,
    new_environment_path: str,
    prerequisites: EnvironmentSpecPrerequisites,
) -> None:
    """
    Creates a pip environment in new_environment_path. requirements_file_path should
    point to a requirements.txt file that will determine what packages are installed.

    There should usually be a filelock around this function.
    """
    # this code is roughly equivalent to the code in PipDockerfile and PipDockerfile
    return_code = await (
        await asyncio.create_subprocess_exec(
            "python", "-m", "venv", new_environment_path
        )
    ).wait()
    if return_code != 0:
        raise ValueError(
            f"venv creation in {new_environment_path} failed with return code "
            f"{return_code}"
        )

    disable_pip_version_check = {"PIP_DISABLE_PIP_VERSION_CHECK": "1"}
    pip_install = [
        os.path.join(new_environment_path, "bin", "python"),
        "-m",
        "pip",
        "install",
    ]

    if prerequisites & EnvironmentSpecPrerequisites.GOOGLE_AUTH:
        return_code = await (
            await asyncio.create_subprocess_exec(
                *pip_install,
                GOOGLE_AUTH_PACKAGE,
                env=disable_pip_version_check,
            )
        ).wait()
        if return_code != 0:
            raise ValueError(
                f"Installing pre-requirements for {requirements_file_path} in "
                f"{new_environment_path} failed with return code {return_code}"
            )

    return_code = await (
        await asyncio.create_subprocess_exec(
            *pip_install,
            "-r",
            requirements_file_path,
            env=disable_pip_version_check,
        )
    ).wait()
    if return_code != 0:
        raise ValueError(
            f"Installing requirements from {requirements_file_path} in "
            f"{new_environment_path} failed with return code {return_code}"
        )



from __future__ import annotations

import copy
import os
import os.path
import pickle
import shlex
import shutil
import sys
import urllib.parse
import uuid
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Optional,
    Sequence,
    TYPE_CHECKING,
    Tuple,
    TypeVar,
    Union,
)

import cloudpickle

from meadowrun.config import JOB_ID_VALID_CHARACTERS, MEADOWRUN_INTERPRETER
from meadowrun.deployment_spec import (
    ContainerAtDigestInterpreter,
    ContainerInterpreter,
    ContainerInterpreterBase,
    Deployment,
)
from meadowrun.docker_controller import get_registry_domain
from meadowrun.meadowrun_pb2 import (
    AwsSecretProto,
    AzureSecretProto,
    CodeZipFile,
    ContainerAtDigest,
    ContainerAtTag,
    ContainerImage,
    Credentials,
    CredentialsSourceMessage,
    EnvironmentSpec,
    EnvironmentSpecInCode,
    GitRepoBranch,
    GitRepoCommit,
    Job,
    KubernetesSecretProto,
    PyCommandJob,
    PyFunctionJob,
    QualifiedFunctionName,
    ServerAvailableContainer,
    ServerAvailableFile,
    ServerAvailableFolder,
    ServerAvailableInterpreter,
    StringPair,
)
from meadowrun.run_job_core import (
    Host,
    JobCompletion,
    Resources,
    TaskResult,
    WaitOption,
)
from meadowrun.storage_grid_job import ensure_uploaded_incremental
from meadowrun.storage_keys import STORAGE_CODE_CACHE_PREFIX

if TYPE_CHECKING:
    from meadowrun.deployment_internal_types import (
        CodeDeployment,
        InterpreterDeployment,
        VersionedCodeDeployment,
        VersionedInterpreterDeployment,
    )
    from meadowrun.credentials import CredentialsSourceForService

_T = TypeVar("_T")
_U = TypeVar("_U")


def _credentials_source_message(
    credentials_source: CredentialsSourceForService,
) -> CredentialsSourceMessage:
    result = CredentialsSourceMessage(
        service=Credentials.Service.Value(credentials_source.service),
        service_url=credentials_source.service_url,
    )
    if isinstance(credentials_source.source, AwsSecretProto):
        result.aws_secret.CopyFrom(credentials_source.source)
    elif isinstance(credentials_source.source, AzureSecretProto):
        result.azure_secret.CopyFrom(credentials_source.source)
    elif isinstance(credentials_source.source, ServerAvailableFile):
        result.server_available_file.CopyFrom(credentials_source.source)
    elif isinstance(credentials_source.source, KubernetesSecretProto):
        result.kubernetes_secret.CopyFrom(credentials_source.source)
    else:
        raise ValueError(
            f"Unknown type of credentials source {type(credentials_source.source)}"
        )
    return result


async def _get_job_fields(
    host: Host,
    resources: Resources,
    deployment: Union[Deployment, Awaitable[Deployment], None],
    sidecar_containers: Union[
        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
    ],
) -> Dict[str, Any]:
    """
    Compiles deployment and sidecar_containers into protobuf messages that can be added
    to the Job object. Should be used like Job(**(await _get_job_fields(...)))
    """

    # this doesn't totally belong here but it's close enough
    await host.set_defaults()

    # after this deployment will always be Deployment
    if deployment is None:
        deployment = await Deployment.mirror_local()
    elif not isinstance(deployment, Deployment):
        # TODO run this in parallel with e.g. launching instances
        deployment = await deployment

    # populate credentials_sources part 1
    credentials_sources = [
        _credentials_source_message(c) for c in deployment.credentials_sources
    ]

    # set environment_variables
    if deployment.environment_variables:
        environment_variables = [
            StringPair(key=key, value=value)
            for key, value in deployment.environment_variables.items()
        ]
    else:
        environment_variables = []

    # set interpreter
    interpreter = deployment.interpreter or ServerAvailableInterpreter(
        interpreter_path=MEADOWRUN_INTERPRETER
    )
    if resources.needs_cuda() and isinstance(
        interpreter, (EnvironmentSpec, EnvironmentSpecInCode)
    ):
        # we don't want to modify the interpreter in place
        interpreter = copy.deepcopy(interpreter)
        interpreter.additional_software["cuda"] = ""

    # set code
    code = deployment.code or ServerAvailableFolder()
    if isinstance(code, CodeZipFile):
        # we don't want to modify the CodeZipFile in place
        code = copy.deepcopy(code)
        await _upload_code_zip_file(code, host)

    # prepare sidecar_containers
    (
        sidecar_containers_prepared,
        additional_credentials_sources,
    ) = _prepare_sidecar_containers(sidecar_containers)
    credentials_sources.extend(additional_credentials_sources)

    return {
        "result_highest_pickle_protocol": pickle.HIGHEST_PROTOCOL,
        "environment_variables": environment_variables,
        "sidecar_containers": sidecar_containers_prepared,
        "credentials_sources": credentials_sources,
        "uses_gpu": resources.uses_gpu(),
        _job_field_for_code_deployment(code): code,
        _job_field_for_interpreter_deployment(interpreter): interpreter,
    }


def _file_path_from_url(file_url: str) -> str:
    decoded_url = urllib.parse.urlparse(file_url)
    if decoded_url.scheme != "file":
        raise ValueError(f"Expected file URI: {file_url}")
    if sys.platform == "win32" and decoded_url.path.startswith("/"):
        # on Windows, file:///C:\foo turns into file_url.path = /C:\foo so we need
        # to remove the forward slash at the beginning
        file_path = decoded_url.path[1:]
    else:
        file_path = decoded_url.path
    return file_path


async def _upload_code_zip_file(code_deploy: CodeZipFile, host: Host) -> None:
    """
    Modifies code_deploy in place!

    CodeZipFile is not usable as is. mirror_local produces a CodeZipFile where the "url"
    is on the local file system. mirror_local doesn't have access to a
    host/storage_bucket, so it can't do the upload, so we "finish off" that part here
    """
    async with await host.get_storage_bucket() as storage_bucket:
        file_url = code_deploy.url
        file_path = _file_path_from_url(file_url)

        key = await ensure_uploaded_incremental(
            storage_bucket, file_path, STORAGE_CODE_CACHE_PREFIX
        )
        # this "url" is a bit silly, we just want to be able to distinguish from a
        # file url. The only information we really care about is the key
        code_deploy.url = urllib.parse.urlunparse(("mdrstorage", "_", key, "", "", ""))
        # TODO this won't work if we ever want to reuse the output of a mirror_local
        shutil.rmtree(os.path.dirname(file_path), ignore_errors=True)


def _pickle_protocol_for_deployed_interpreter() -> int:
    """
    This is a placeholder, the intention is to get the deployed interpreter's version
    somehow from the Deployment object or something like it and use that to determine
    what the highest pickle protocol version we can use safely is.
    """

    # TODO just hard-coding the interpreter version for now, need to actually grab it
    #  from the deployment somehow
    interpreter_version = (3, 8, 0)

    # based on documentation in
    # https://docs.python.org/3/library/pickle.html#data-stream-format
    if interpreter_version >= (3, 8, 0):
        protocol = 5
    elif interpreter_version >= (3, 4, 0):
        protocol = 4
    elif interpreter_version >= (3, 0, 0):
        protocol = 3
    else:
        # TODO support for python 2 would require dealing with the string/bytes issue
        raise NotImplementedError("We currently only support python 3")

    return min(protocol, pickle.HIGHEST_PROTOCOL)


def _make_valid_friendly_name(job_id: str) -> str:
    return "".join(c for c in job_id if c in JOB_ID_VALID_CHARACTERS)


def _get_friendly_name(function: Callable[[_T], _U]) -> str:
    friendly_name = getattr(function, "__name__", "")
    if not friendly_name:
        friendly_name = "lambda"

    return _make_valid_friendly_name(friendly_name)


def _prepare_ports(
    ports: Union[Iterable[str], str, Iterable[int], int, None]
) -> Optional[Sequence[str]]:
    if ports is None:
        return None
    elif isinstance(ports, int):
        return [str(ports)]
    elif isinstance(ports, str):
        return [ports]
    else:
        return [str(p) for p in ports]


def _prepare_sidecar_containers(
    container_interpreters: Union[
        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
    ] = None,
) -> Tuple[Sequence[ContainerImage], Sequence[CredentialsSourceMessage]]:
    if container_interpreters is None:
        return [], []

    if isinstance(container_interpreters, ContainerInterpreterBase):
        container_interpreters = [container_interpreters]

    sidecar_containers = []
    credentials_sources = []
    for interpreter in container_interpreters:
        if isinstance(interpreter, ContainerAtDigestInterpreter):
            sidecar_containers.append(
                ContainerImage(
                    container_image_at_digest=ContainerAtDigest(
                        repository=interpreter.repository_name,
                        digest=interpreter.digest,
                    )
                )
            )
        elif isinstance(interpreter, ContainerInterpreter):
            sidecar_containers.append(
                ContainerImage(
                    container_image_at_tag=ContainerAtTag(
                        repository=interpreter.repository_name,
                        tag=interpreter.tag,
                    )
                )
            )
        else:
            raise ValueError(
                f"Unexpected type of sidecar_container {type(interpreter)}"
            )

        username_password_secret = interpreter._get_username_password_secret()
        if username_password_secret is not None:
            credentials_sources.append(
                _credentials_source_message(
                    username_password_secret._to_credentials_source(
                        "DOCKER",
                        get_registry_domain(interpreter._get_repository_name())[0],
                        Credentials.Type.USERNAME_PASSWORD,
                    )
                )
            )

    return sidecar_containers, credentials_sources


async def run_function(
    function: Union[Callable[..., _T], str],
    host: Host,
    resources: Optional[Resources] = None,
    deployment: Union[Deployment, Awaitable[Deployment], None] = None,
    args: Optional[Sequence[Any]] = None,
    kwargs: Optional[Dict[str, Any]] = None,
    sidecar_containers: Union[
        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
    ] = None,
    ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
    wait_for_result: bool = True,
) -> _T:
    """
    Runs function on a remote machine, specified by "host".

    Args:
        function: A reference to a function (e.g. `package.module.function_name`), a
            lambda, or a string like `"package.module.function_name"` (which is useful
            if the function cannot be referenced in the current environment but can be
            referenced in the deployed environment)
        host: Specifies where to run the function. See [Host][meadowrun.Host] and
            derived classes.
        resources: Specifies the resources (e.g. CPU, RAM) needed by the
            function. For some hosts, this is optional, for other hosts it is required.
            See [Resources][meadowrun.Resources].
        deployment: See [Deployment][meadowrun.Deployment]. Specifies the environment
            (code and libraries) that are needed to run this command. This can be an
            actual Deployment object, or it can be an Awaitable that will produce a
            Deployment object. The default, None, is equivalent to
            [mirror_local][meadowrun.Deployment.mirror_local]
        args: Passed to the function like `function(*args)`
        kwargs: Passed to the function like `function(**kwargs)`
        sidecar_containers: Additional containers that will be available from the main
            job as sidecar-container-0 sidecar-container-1, etc.
        ports: A specification of ports to make available on the machine that runs this
            job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened just for the
            duration of this job. Be careful as other jobs could be running on the same
            machine at the same time!
        wait_for_result: If this is set to False, we will run in "fire and forget" mode,
            which kicks off the function and doesn't wait for it to return.

    Returns:
        If wait_for_result is True (which is the default), the return value will be the
            result of calling `function`. If wait_for_result is False, the return value
            will always be None.
    """
    # first, construct the PyFunctionJob

    pickle_protocol = _pickle_protocol_for_deployed_interpreter()

    # TODO add support for compressions, pickletools.optimize, possibly cloudpickle?
    # TODO also add the ability to write this to a shared location so that we don't need
    #  to pass it through the server.
    if args or kwargs:
        pickled_function_arguments = pickle.dumps(
            (args, kwargs), protocol=pickle_protocol
        )
    else:
        # according to docs, None is translated to empty anyway
        pickled_function_arguments = b""

    if isinstance(function, str):
        friendly_name = function
        module_name, separator, function_name = function.rpartition(".")
        if not separator:
            raise ValueError(
                f"Function must be in the form module_name.function_name: {function}"
            )
        py_function = PyFunctionJob(
            pickled_function_arguments=pickled_function_arguments,
            qualified_function_name=QualifiedFunctionName(
                module_name=module_name,
                function_name=function_name,
            ),
        )
    else:
        friendly_name = _get_friendly_name(function)
        pickled_function = cloudpickle.dumps(function)
        # TODO larger functions should get copied to S3/filesystem instead of sent
        # directly
        print(f"Size of pickled function is {len(pickled_function)}")
        py_function = PyFunctionJob(
            pickled_function_arguments=pickled_function_arguments,
            pickled_function=pickled_function,
        )

    # now create the Job and run it
    if resources is None:
        resources = Resources()

    job = Job(
        job_id=str(uuid.uuid4()),
        job_friendly_name=friendly_name,
        py_function=py_function,
        ports=_prepare_ports(ports),
        **(await _get_job_fields(host, resources, deployment, sidecar_containers)),
    )

    job_completion = await host.run_job(
        resources.to_internal(),
        job,
        WaitOption.WAIT_AND_TAIL_STDOUT if wait_for_result else WaitOption.DO_NOT_WAIT,
    )
    return job_completion.result


async def run_command(
    args: Union[str, Sequence[str]],
    host: Host,
    resources: Optional[Resources] = None,
    deployment: Union[Deployment, Awaitable[Deployment], None] = None,
    context_variables: Optional[Dict[str, Any]] = None,
    sidecar_containers: Union[
        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
    ] = None,
    ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
    wait_for_result: bool = True,
) -> JobCompletion[None]:
    """
    Runs the specified command on a remote machine

    Args:
        args: Specifies the command to run, can be a string (e.g. `"jupyter nbconvert
            --to html analysis.ipynb"`) or a list of strings (e.g. `["jupyter",
            --"nbconvert", "--to", "html", "analysis.ipynb"]`)
        host: Specifies where to run the function. See [Host][meadowrun.Host] and
            derived classes.
        resources: Specifies the resources (e.g. CPU, RAM) needed by the
            command. For some hosts, this is optional, for other hosts it is required.
            See [Resources][meadowrun.Resources].
        deployment: See [Deployment][meadowrun.Deployment]. Specifies the environment
            (code and libraries) that are needed to run this command. This can be an
            actual Deployment object, or it can be an Awaitable that will produce a
            Deployment object. The default, None, is equivalent to
            [mirror_local][meadowrun.Deployment.mirror_local]
        context_variables: Experimental feature
        sidecar_containers: Additional containers that will be available from the main
            job as sidecar-container-0 sidecar-container-1, etc.
        ports: A specification of ports to make available on the machine that runs this
            job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened just for the
            duration of this job. Be careful as other jobs could be running on the same
            machine at the same time!
        wait_for_result: If this is set to False, we will run in "fire and forget" mode,
            which kicks off the command and doesn't wait for it to return.

    Returns:
        A JobCompletion object that contains metadata about the running of the job.
    """

    # construct the PyCommandJob

    if isinstance(args, str):
        args = shlex.split(args)
    # this is kind of a silly way to get a friendly name--treat the first three
    # elements of args as if they're paths and take the last part of each path
    friendly_name = "-".join(os.path.basename(arg) for arg in args[:3])

    if context_variables:
        pickled_context_variables = pickle.dumps(
            context_variables, protocol=_pickle_protocol_for_deployed_interpreter()
        )
    else:
        pickled_context_variables = b""

    py_command = PyCommandJob(
        command_line=args, pickled_context_variables=pickled_context_variables
    )

    # create the job and run it

    if resources is None:
        resources = Resources()
    job = Job(
        job_id=str(uuid.uuid4()),
        job_friendly_name=_make_valid_friendly_name(friendly_name),
        py_command=py_command,
        ports=_prepare_ports(ports),
        **(await _get_job_fields(host, resources, deployment, sidecar_containers)),
    )

    return await host.run_job(
        resources.to_internal(),
        job,
        WaitOption.WAIT_AND_TAIL_STDOUT if wait_for_result else WaitOption.DO_NOT_WAIT,
    )


async def run_map(
    function: Callable[[_T], _U],
    args: Sequence[_T],
    host: Host,
    resources_per_task: Optional[Resources] = None,
    deployment: Union[Deployment, Awaitable[Deployment], None] = None,
    num_concurrent_tasks: Optional[int] = None,
    sidecar_containers: Union[
        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
    ] = None,
    ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
    wait_for_result: bool = True,
    max_num_task_attempts: int = 1,
    retry_with_more_memory: bool = False,
) -> Optional[Sequence[_U]]:
    """
    Equivalent to `map(function, args)`, but runs distributed and in parallel.

    Args:
        function: A reference to a function (e.g. `package.module.function_name`) or a
            lambda
        args: A list of objects, each item in the list represents a "task",
            where each "task" is an invocation of `function` on the item in the list
        resources_per_task: The resources (e.g. CPU and RAM) required to run a
            single task. For some hosts, this is optional, for other hosts it is
            required. See [Resources][meadowrun.Resources].
        host: Specifies where to get compute resources from. See [Host][meadowrun.Host]
            and derived classes.
        num_concurrent_tasks: The number of workers to launch. This can be less than or
            equal to the number of args/tasks. Will default to half the total number of
            tasks plus one, rounded down if set to None.
        deployment: See [Deployment][meadowrun.Deployment]. Specifies the environment
            (code and libraries) that are needed to run this command. This can be an
            actual Deployment object, or it can be an Awaitable that will produce a
            Deployment object. The default, None, is equivalent to
            [mirror_local][meadowrun.Deployment.mirror_local]
        sidecar_containers: Additional containers that will be available from the main
            job as sidecar-container-0 sidecar-container-1, etc.
        ports: A specification of ports to make available on the machines that runs
            tasks for this job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened
            just for the duration of this job. Be careful as other jobs could be running
            on the same machine at the same time!
        wait_for_result: If this is set to False, we will run in "fire and forget" mode,
            which kicks off the command and doesn't wait for it to return.
        max_num_task_attempts: If this is set to more than 1, tasks that fail will be
            retried. If this parameter is e.g. 3, a task that fails will be attempted a
            total of 3 times.
        retry_with_more_memory: This is an experimental feature and the API will likely
            change. If this is set to True, when a task fails, if the task at some point
            used more than 95% of the requested memory, the task will be retried with
            more memory. Each attempt will be allocated (original requested memory) *
            (attempt number).

    Returns:
        If wait_for_result is True (which is the default), the return value will be the
            result of running `function` on each of `args`. If wait_for_result is False,
            the return value will always be None.
    """

    if not num_concurrent_tasks:
        num_concurrent_tasks = len(args) // 2 + 1
    else:
        num_concurrent_tasks = min(num_concurrent_tasks, len(args))

    # prepare some variables for constructing the worker jobs

    pickle_protocol = _pickle_protocol_for_deployed_interpreter()

    if resources_per_task is None:
        resources_per_task = Resources()
    job_fields = await _get_job_fields(
        host, resources_per_task, deployment, sidecar_containers
    )
    job_fields["job_friendly_name"] = _get_friendly_name(function)
    job_fields["ports"] = _prepare_ports(ports)

    if not wait_for_result:
        wait_option = WaitOption.DO_NOT_WAIT
    elif num_concurrent_tasks == 1:
        wait_option = WaitOption.WAIT_AND_TAIL_STDOUT
    else:
        wait_option = WaitOption.WAIT_SILENTLY

    return await host.run_map(
        function,
        args,
        resources_per_task.to_internal(),
        job_fields,
        num_concurrent_tasks,
        pickle_protocol,
        wait_option,
        max_num_task_attempts,
        retry_with_more_memory,
    )


async def run_map_as_completed(
    function: Callable[[_T], _U],
    args: Sequence[_T],
    host: Host,
    resources_per_task: Optional[Resources] = None,
    deployment: Union[Deployment, Awaitable[Deployment], None] = None,
    num_concurrent_tasks: Optional[int] = None,
    sidecar_containers: Union[
        Iterable[ContainerInterpreterBase], ContainerInterpreterBase, None
    ] = None,
    ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
    max_num_task_attempts: int = 1,
    retry_with_more_memory: bool = False,
) -> AsyncIterable[TaskResult[_U]]:
    """
    Equivalent to [run_map][meadowrun.run_map], but returns results from tasks as they
    are completed. This means that to access the results, you need to iterate using
    `async for`, and call `result_or_raise`  on the returned
    [TaskResult][meadowrun.TaskResult] objects. Usage for approximating `run_map`
    behavior is:

    ```python
    sorted_tasks = sorted(
        [task async for task in run_map_as_completed(...)],
        key=lambda t: t.task_id
    )
    results = [task.result_or_raise() for task in sorted_tasks]
    ```

    Args:
        function: A reference to a function (e.g. `package.module.function_name`) or a
            lambda
        args: A list of objects, each item in the list represents a "task",
            where each "task" is an invocation of `function` on the item in the list
        resources_per_task: The resources (e.g. CPU and RAM) required to run a
            single task. For some hosts, this is optional, for other hosts it is
            required. See [Resources][meadowrun.Resources].
        host: Specifies where to get compute resources from. See [Host][meadowrun.Host]
            and derived classes.
        num_concurrent_tasks: The number of workers to launch. This can be less than or
            equal to the number of args/tasks. Will default to half the total number of
            tasks plus one, rounded down if set to None.
        deployment: See [Deployment][meadowrun.Deployment]. Specifies the environment
            (code and libraries) that are needed to be an Awaitable that will produce a
            Deployment object. The default, None, is equivalent to
            [mirror_local][meadowrun.Deployment.mirror_local]
        sidecar_containers: Additional containers that will be available from the main
            job as sidecar-container-0 sidecar-container-1, etc.
        ports: A specification of ports to make available on the machines that runs
            tasks for this job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened
            just for the duration of this job. Be careful as other jobs could be running
            on the same machine at the same time!
        max_num_task_attempts: If this is set to more than 1, tasks that fail will be
            retried. If this parameter is e.g. 3, a task that fails will be attempted a
            total of 3 times.
        retry_with_more_memory: This is an experimental feature and the API will likely
            change. If this is set to True, when a task fails, if the task at some point
            used more than 95% of the requested memory, the task will be retried with
            more memory. Each attempt will be allocated (original requested memory) *
            (attempt number).

    Returns:
        An async iterable returning [TaskResult][meadowrun.TaskResult] objects.
    """

    if not num_concurrent_tasks:
        num_concurrent_tasks = len(args) // 2 + 1
    else:
        num_concurrent_tasks = min(num_concurrent_tasks, len(args))

    # prepare some variables for constructing the worker jobs
    friendly_name = _get_friendly_name(function)

    pickle_protocol = _pickle_protocol_for_deployed_interpreter()

    if resources_per_task is None:
        resources_per_task = Resources()
    job_fields = await _get_job_fields(
        host, resources_per_task, deployment, sidecar_containers
    )
    job_fields["job_friendly_name"] = friendly_name
    job_fields["ports"] = _prepare_ports(ports)

    return host.run_map_as_completed(
        function,
        args,
        resources_per_task.to_internal(),
        job_fields,
        num_concurrent_tasks,
        pickle_protocol,
        wait_for_result=WaitOption.WAIT_SILENTLY,
        max_num_task_attempts=max_num_task_attempts,
        retry_with_more_memory=retry_with_more_memory,
    )


def _job_field_for_code_deployment(
    code_deployment: Union[CodeDeployment, VersionedCodeDeployment],
) -> str:
    """
    You can't just do Job(code_deployment=code_deployment) because of the protobuf
    oneofs. Instead you need to do
    Job(**{_job_field_for_code_deployment(code_deployment): code_deployment})
    """
    if isinstance(code_deployment, ServerAvailableFolder):
        return "server_available_folder"
    elif isinstance(code_deployment, GitRepoCommit):
        return "git_repo_commit"
    elif isinstance(code_deployment, GitRepoBranch):
        return "git_repo_branch"
    elif isinstance(code_deployment, CodeZipFile):
        return "code_zip_file"
    else:
        raise ValueError(f"Unknown code deployment type {type(code_deployment)}")


def _job_field_for_interpreter_deployment(
    interpreter_deployment: Union[InterpreterDeployment, VersionedInterpreterDeployment]
) -> str:
    """See _job_field_for_code_deployment for usage pattern"""
    if isinstance(interpreter_deployment, ServerAvailableInterpreter):
        return "server_available_interpreter"
    elif isinstance(interpreter_deployment, ContainerAtDigest):
        return "container_at_digest"
    elif isinstance(interpreter_deployment, ServerAvailableContainer):
        return "server_available_container"
    elif isinstance(interpreter_deployment, ContainerAtTag):
        return "container_at_tag"
    elif isinstance(interpreter_deployment, EnvironmentSpecInCode):
        return "environment_spec_in_code"
    elif isinstance(interpreter_deployment, EnvironmentSpec):
        return "environment_spec"
    else:
        raise ValueError(
            f"Unknown interpreter deployment type {type(interpreter_deployment)}"
        )



"""
This code cannot reference anything outside of mgmt_functions (as that's what gets
uploaded to the Azure function). We use relative imports which will work both in the
"regular" environment as well as in the Azure function
"""

import asyncio
import datetime
import logging
import os
from typing import Any, Dict, List

import aiohttp

from ..azure_core.azure_exceptions import (
    AzureRestApiError,
)

from ..azure_core.azure_acr import get_tags_in_repository, delete_tag
from ..azure_core.azure_rest_api import (
    azure_rest_api_paged,
    azure_rest_api,
    parse_azure_timestamp,
)
from ..azure_core.azure_storage_api import (
    StorageAccount,
    azure_table_api,
    azure_table_api_paged,
    table_key_url,
)
from ..azure_constants import (
    CONTAINER_IMAGE,
    GRID_TASK_QUEUE,
    LAST_USED_TABLE_NAME,
    MEADOWRUN_SUBSCRIPTION_ID,
    QUEUE_NAME_TIMESTAMP_FORMAT,
    _MEADOWRUN_GENERATED_DOCKER_REPO,
    _REQUEST_QUEUE_NAME_PREFIX,
    _RESULT_QUEUE_NAME_PREFIX,
    RESOURCE_TYPES_TYPE,
    meadowrun_container_registry_name,
)
from ..mgmt_functions_shared import get_storage_account, get_resource_group_path

# a queue that has been inactive for this time will get cleaned up
_QUEUE_INACTIVE_TIME = datetime.timedelta(hours=4)
# a container image has not been used for this time will get cleaned up
_CONTAINER_IMAGE_UNUSED_TIME = datetime.timedelta(days=4)


async def _get_last_used_records(
    storage_account: StorageAccount, resource_type: RESOURCE_TYPES_TYPE
) -> Dict[Any, datetime.datetime]:
    try:
        last_used_records = {
            item["RowKey"]: parse_azure_timestamp(item["Timestamp"])
            async for page in azure_table_api_paged(
                "GET",
                storage_account,
                f"{LAST_USED_TABLE_NAME}()",
                query_parameters={"$filter": f"PartitionKey eq '{resource_type}'"},
            )
            for item in page["value"]
        }
        return last_used_records
    except AzureRestApiError as exn:
        if exn.status != 404 or exn.code != "TableNotFound":
            raise
        return {}


async def delete_old_task_queues() -> List[str]:
    """See _deregister_and_terminate_vms for why we return log statements"""
    logs = []

    storage_account = get_storage_account()
    resource_group_path = get_resource_group_path()

    # the last used records are keyed off of the job_id for the queue, whereas the queue
    # names are in the form of {prefix}-{job_id}-{created_timestamp}

    now = datetime.datetime.utcnow()
    delete_tasks = []

    last_used_records = await _get_last_used_records(storage_account, GRID_TASK_QUEUE)

    surviving_job_ids = set()
    deleted_job_ids = set()

    queues_path = (
        f"{resource_group_path}/providers/Microsoft.Storage/storageAccounts"
        f"/{storage_account.name}/queueServices/default/queues"
    )

    async for page in azure_rest_api_paged("GET", queues_path, "2021-09-01"):
        for queue in page["value"]:
            queue_name = queue["name"]
            if not queue_name.startswith(
                _REQUEST_QUEUE_NAME_PREFIX
            ) and not queue_name.startswith(_RESULT_QUEUE_NAME_PREFIX):
                # this is not a meadowrun grid task queue
                continue

            # first parse the queue names, while deleting any queues that don't fit the
            # expected patterns

            prefix, sep, remainder = queue_name.partition("-")
            job_id, sep, created_timestamp_string = remainder.rpartition("-")
            if sep != "-":
                logs.append(
                    "Queue name was not in the expected prefix-job_id-timestamp format:"
                    f" {queue_name}, deleting it"
                )
                delete_tasks.append(
                    asyncio.create_task(
                        azure_rest_api(
                            "DELETE", f"{queues_path}/{queue_name}", "2021-09-01"
                        )
                    )
                )
                continue

            # next, if we have a last-used record that's too old, delete the queue,
            # otherwise mark it as a "surviving" queue so that we know to keep the
            # last_used record around
            if job_id in last_used_records:
                last_used = last_used_records[job_id]
                if now - last_used > _QUEUE_INACTIVE_TIME:
                    logs.append(
                        f"Queue {queue_name} was last used at {last_used}, deleting"
                    )
                    deleted_job_ids.add(job_id)
                    delete_tasks.append(
                        asyncio.create_task(
                            azure_rest_api(
                                "DELETE", f"{queues_path}/{queue_name}", "2021-09-01"
                            )
                        )
                    )
                else:
                    surviving_job_ids.add(job_id)
                continue

            # finally, we don't have any last-used records, so we have to use the
            # created timestamp
            try:
                created_timestamp = datetime.datetime.strptime(
                    created_timestamp_string, QUEUE_NAME_TIMESTAMP_FORMAT
                )
            except ValueError:
                logs.append(
                    f"Queue name {queue_name} is in the format prefix-job_id-timestamp,"
                    " but the timestamp cannot be parsed, deleting the queue"
                )
                delete_tasks.append(
                    asyncio.create_task(
                        azure_rest_api(
                            "DELETE", f"{queues_path}/{queue_name}", "2021-09-01"
                        )
                    )
                )
                continue

            if now - created_timestamp > _QUEUE_INACTIVE_TIME:
                logs.append(
                    f"Queue {queue_name} has no last used records and was created at "
                    f"{created_timestamp}, deleting"
                )
                delete_tasks.append(
                    asyncio.create_task(
                        azure_rest_api(
                            "DELETE", f"{queues_path}/{queue_name}", "2021-09-01"
                        )
                    )
                )
                continue

    # now delete last_used records that don't correspond to any existing queues
    for job_id in last_used_records.keys():
        if job_id in surviving_job_ids:
            continue

        if job_id not in deleted_job_ids:
            logs.append(
                f"job_id {job_id} has a last_used record, but no existing queues, "
                "deleting the last_used record now"
            )

        # if we did delete the corresponding queue, still delete the last_used
        # record, just no need to log
        delete_tasks.append(
            asyncio.create_task(
                azure_table_api(
                    "DELETE",
                    storage_account,
                    table_key_url(LAST_USED_TABLE_NAME, GRID_TASK_QUEUE, job_id),
                    additional_headers={"If-Match": "*"},
                )
            )
        )

    if delete_tasks:
        await asyncio.wait(delete_tasks)

    return logs


async def delete_unused_images() -> List[str]:
    """See _deregister_and_terminate_vms for why we return log statements"""
    logs = []

    storage_account = get_storage_account()

    registry_name = meadowrun_container_registry_name(
        os.environ[MEADOWRUN_SUBSCRIPTION_ID]
    )

    delete_tasks = []

    now = datetime.datetime.now()

    last_used_records = await _get_last_used_records(storage_account, CONTAINER_IMAGE)

    deleted_tags = set()
    surviving_tags = set()

    try:
        tags = await get_tags_in_repository(
            registry_name, _MEADOWRUN_GENERATED_DOCKER_REPO
        )
    except aiohttp.ClientConnectorError:
        # assume the registry does not exist => no tags
        tags = []

    for tag in tags:
        tag_name = tag["name"]

        # first see if we have a last used record for this tag
        if tag_name in last_used_records:
            last_used = last_used_records[tag_name]
            if now - last_used > _CONTAINER_IMAGE_UNUSED_TIME:
                logs.append(
                    f"Image {tag_name} will be deleted, was last used at {last_used}"
                )
                delete_tasks.append(
                    asyncio.create_task(
                        delete_tag(
                            registry_name, _MEADOWRUN_GENERATED_DOCKER_REPO, tag_name
                        )
                    )
                )
                deleted_tags.add(tag_name)
            else:
                surviving_tags.add(tag_name)
            continue

        # if we don't have a last used record, use the last_updated property
        if (
            now - parse_azure_timestamp(tag["lastUpdateTime"])
            > _CONTAINER_IMAGE_UNUSED_TIME
        ):
            logs.append(
                f"Image {tag_name} will be deleted, has not been used and last "
                f"updated at {tag['lastUpdateTime']}"
            )
            delete_tasks.append(
                asyncio.create_task(
                    delete_tag(
                        registry_name, _MEADOWRUN_GENERATED_DOCKER_REPO, tag_name
                    )
                )
            )

    for tag_name in last_used_records.keys():
        if tag_name in surviving_tags:
            continue

        if tag_name not in deleted_tags:
            logs.append(
                f"Image {tag_name} has a last_used record but the image does not "
                "exist. Deleting the last_used record now"
            )
        # if we did delete the corresponding image, still delete the last_used
        # record, just no need to log
        delete_tasks.append(
            asyncio.create_task(
                azure_table_api(
                    "DELETE",
                    storage_account,
                    table_key_url(LAST_USED_TABLE_NAME, CONTAINER_IMAGE, tag_name),
                    additional_headers={"If-Match": "*"},
                )
            )
        )

    if delete_tasks:
        await asyncio.wait(delete_tasks)

    return logs


def main(myTimer) -> None:  # type: ignore
    """
    The entry point for the Azure function (see function.json). Configured to run every
    minute by default.

    myTime should be annotated as azure.functions.TimerRequest, but we'd rather not add
    that dependency (it exists by default in the function runtime where this function
    will actually run). Also, the variable cannot be snake_case, and can only be changed
    if the corresponding name in function.json is changed.
    """
    logs = asyncio.run(delete_old_task_queues())
    for log in logs:
        logging.info(log)
    logs = asyncio.run(delete_unused_images())
    for log in logs:
        logging.info(log)



"""
Constants needed both by mgmt_functions and the "regular" code outside of
mgmt_functions.
"""

from __future__ import annotations

from typing_extensions import Literal

# this should almost never be used directly instead of ensure_meadowrun_resource_group.
MEADOWRUN_RESOURCE_GROUP_NAME = "Meadowrun-rg"

VM_ALLOC_TABLE_NAME = "meadowrunVmAlloc"
LAST_USED_TABLE_NAME = "meadowrunLastUsed"

RESOURCE_TYPES_TYPE = Literal["GRID_TASK_QUEUE", "CONTAINER_IMAGE"]
GRID_TASK_QUEUE: RESOURCE_TYPES_TYPE = "GRID_TASK_QUEUE"
CONTAINER_IMAGE: RESOURCE_TYPES_TYPE = "CONTAINER_IMAGE"

_REQUEST_QUEUE_NAME_PREFIX = "mrgtrequest"
_RESULT_QUEUE_NAME_PREFIX = "mrgtresult"

QUEUE_NAME_TIMESTAMP_FORMAT = "%Y%m%d%H%M%S"

ALLOCATED_TIME = "allocated_time"
JOB_ID = "job_id"
LAST_UPDATE_TIME = "last_update_time"
RESOURCES_AVAILABLE = "resources_available"
NON_CONSUMABLE_RESOURCES = "non_consumable_resources"
RESOURCES_ALLOCATED = "resources_allocated"
RUNNING_JOBS = "running_jobs"
PREVENT_FURTHER_ALLOCATION = "prevent_further_allocation"
SINGLE_PARTITION_KEY = "_"
PUBLIC_ADDRESS = "public_address"


MEADOWRUN_STORAGE_ACCOUNT_VARIABLE = "MEADOWRUN_STORAGE_ACCOUNT"
MEADOWRUN_STORAGE_ACCOUNT_KEY_VARIABLE = "MEADOWRUN_STORAGE_ACCOUNT_KEY"
MEADOWRUN_SUBSCRIPTION_ID = "MEADOWRUN_SUBSCRIPTION_ID"

_MEADOWRUN_GENERATED_DOCKER_REPO = "meadowrun_generated"


def meadowrun_container_registry_name(subscription_id: str) -> str:
    """
    Azure Container registry names must be 5 to 50 characters, alphanumeric, and
    globally unique.
    """
    return "mr" + subscription_id.replace("-", "")



from typing import Dict, Tuple, Optional, Iterable, Any, List, cast

import aiohttp

from meadowrun.azure_integration.azure_meadowrun_core import get_subscription_id
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api_paged,
)
from meadowrun.config import (
    GPU,
    EVICTION_RATE_INVERSE,
    LOGICAL_CPU,
    MEMORY_GB,
)
from meadowrun.instance_selection import (
    CloudInstanceType,
    ON_DEMAND_OR_SPOT_VALUES,
    OnDemandOrSpotType,
    ResourcesInternal,
)

_RELEVANT_CAPABILITIES = {
    "GPUs",
    "HyperVGenerations",
    "MemoryGB",
    "vCPUs",
    "vCPUsAvailable",
}


async def _get_vm_skus(location: str) -> Iterable[Tuple[str, Dict[str, float]]]:
    """Returns [(name, consumable_resources), ...]"""
    seen_names = set()
    result = []

    # Unfortunately the slightly friendlier client.virtual_machine_sizes.list function
    # is deprecated as per
    # https://docs.microsoft.com/en-us/rest/api/compute/virtual-machine-sizes/list
    #
    # The new API is https://docs.microsoft.com/en-us/rest/api/compute/resource-skus The
    # underlying REST API does not support any filters other than location
    async for page in azure_rest_api_paged(
        "GET",
        f"subscriptions/{await get_subscription_id()}/providers/Microsoft.Compute/skus",
        "2021-07-01",
        query_parameters={"$filter": f"location eq '{location}'"},
    ):
        for r in page["value"]:
            if r["resourceType"] == "virtualMachines":
                capabilities = {
                    c["name"]: c["value"]
                    for c in r["capabilities"]
                    if c["name"] in _RELEVANT_CAPABILITIES
                }
                if "V2" in capabilities.get("HyperVGenerations", ""):
                    consumable_resources: Dict[str, float] = {}

                    # can't find documentation on the difference between vCPUsAvailable
                    # and vCPUs. When they are different, vCPUsAvailable seems to match
                    # what appears on the Azure Portal. vCPUsAvailable is not always
                    # available, though, in that case vCPUs seems to be accurate, but
                    # have not checked this exhaustively
                    logical_cpu_str = capabilities.get("vCPUsAvailable")
                    if logical_cpu_str is None:
                        logical_cpu_str = capabilities.get("vCPUs")
                    try:
                        consumable_resources[LOGICAL_CPU] = float(
                            cast(str, logical_cpu_str)
                        )
                    except ValueError:
                        print(
                            f"Warning {r['name']} could not be processed because vCPUs "
                            "is missing or cannot be parsed to a float "
                            f"{logical_cpu_str}"
                        )
                        break

                    memory_gb_str = capabilities.get("MemoryGB")
                    try:
                        consumable_resources[MEMORY_GB] = float(
                            cast(str, memory_gb_str)
                        )
                    except ValueError:
                        print(
                            f"Warning {r['name']} could not be processed because "
                            "MemoryGB is missing or cannot be parsed to a float "
                            f"{memory_gb_str}"
                        )
                        break

                    if "GPUs" in capabilities:
                        try:
                            consumable_resources[GPU] = float(capabilities["GPUs"])
                        except ValueError:
                            print(
                                f"Warning ignoring GPUs available on {r['name']} "
                                "because the field could not be parsed to a float "
                                f"{capabilities['GPUs']}"
                            )

                    if r["name"] in seen_names:
                        print(f"Warning skipping duplicate entry for {r['name']}")
                        break

                    seen_names.add(r["name"])
                    result.append((r["name"], consumable_resources))

    return result


async def _get_vm_prices(
    location: str,
) -> Dict[Tuple[str, OnDemandOrSpotType], Tuple[float, Optional[float]]]:
    """
    Returns (VM SKU, on_demand_or_spot) -> (price per hour, None). The None is there so
    that this can be used as a drop-in replacement for get_cached_vm_prices_and_eviction
    if the cached data is not available (in that function, the None would be the
    eviction rate)
    """
    all_prices: Dict[Tuple[str, OnDemandOrSpotType], Tuple[float, Optional[float]]] = {}

    # https://docs.microsoft.com/en-us/rest/api/cost-management/retail-prices/azure-retail-prices
    _initial_prices_url = (
        "https://prices.azure.com/api/retail/prices?$filter="
        f"serviceName eq 'Virtual Machines' and armRegionName eq '{location}' "
        # other options here are DevTestConsumption for
        # https://azure.microsoft.com/en-us/pricing/dev-test/#overview and Reservation
        # for reserved VMs
        "and priceType eq 'Consumption'"
    )

    async with aiohttp.request("GET", _initial_prices_url) as response:
        response.raise_for_status()
        response_json = await response.json()
        _add_new_price_records(all_prices, response_json["Items"])
        next_page_link = response_json.get("NextPageLink")

    i = 1
    while next_page_link:
        if i % 5 == 0:
            print(f"Getting page {i} of Azure VM prices")
        i += 1
        async with aiohttp.request("GET", next_page_link) as response:
            response.raise_for_status()
            response_json = await response.json()
            _add_new_price_records(all_prices, response_json["Items"])
            next_page_link = response_json.get("NextPageLink")

    return all_prices


def _add_new_price_records(
    all_prices: Dict[Tuple[str, OnDemandOrSpotType], Tuple[float, Optional[float]]],
    new_records: Iterable[Dict[str, Any]],
) -> None:
    """Helper function for _get_vm_prices"""
    for r in new_records:
        # dedicated hosts are not VMs
        if r["productName"].endswith(" Dedicated Host") or r["productName"].endswith(
            " DedicatedHost"
        ):
            continue

        if r["productName"].endswith(" Windows"):
            # TODO eventually we want to support Windows as well
            continue

        if r["currencyCode"] != "USD":
            print(
                f"Warning, skipping price record for {r['armSkuName']} because the "
                f"currencyCode is not USD as expected: {r['currencyCode']}"
            )
            continue
        if r["unitOfMeasure"] != "1 Hour":
            print(
                f"Warning, skipping price record for {r['armSkuName']} because the "
                f"unitOfMeasure is not 1 Hour as expected: {r['unitOfMeasure']}"
            )
            continue

        if r["skuName"].endswith(" Low Priority"):
            # low priority VMs are only available through Azure Batch, basically not
            # relevant
            continue
        elif r["skuName"].endswith(" Spot"):
            on_demand_or_spot: OnDemandOrSpotType = "spot"
        else:
            on_demand_or_spot = "on_demand"

        key = cast(str, r["armSkuName"]), on_demand_or_spot

        if key in all_prices:
            print(f"Warning, got multiple records for {key}")
            continue

        if not isinstance(r["retailPrice"], float) and not isinstance(
            r["retailPrice"], int
        ):
            print(
                f"Warning, skipping price record for {r['armSkuName']} because the "
                f"retailPrice is not a float: {r['retailPrice']}"
            )

        # TODO potentially look at unitPrice as well? According to the docs (link
        # above), retailPrice is "without discount", so presumably unitPrice includes
        # discounts?
        all_prices[key] = r["retailPrice"], None


async def _get_cached_vm_prices_and_eviction(
    location: str,
) -> Dict[Tuple[str, OnDemandOrSpotType], Tuple[float, Optional[float]]]:
    """
    Returns (VM SKU, on_demand_or_spot) -> (price per hour, eviction rate). For eviction
    rate, None means that we do not know the eviction rate, NOT that it is 0. If it is
    0, then eviction rate will be set to 0.

    Tries to first get this data via a pre-cached JSON file stored in an Azure blob (see
    azure_pricing_eviction_cache.py). If that's not available, we'll use the Azure
    pricing APIs which are much slower. We won't have any eviction data unfortunately as
    there is no API.
    """
    async with aiohttp.request(
        "GET",
        (
            "https://meadowrunprod.blob.core.windows.net/azure-pricing/"
            f"{location}-prices-eviction.json"
        ),
    ) as response:
        if response.ok:
            pricing_eviction_data = await response.json()
            return {
                cast(Tuple[str, OnDemandOrSpotType], tuple(key)): value
                for key, value in pricing_eviction_data
            }

    print(
        f"Warning precompiled pricing and eviction rate data for {location} is not "
        "available. Please contact support@meadowdata.io to request an update. We "
        "will proceed with live pricing data (which is slow), and no eviction rate "
        "data (which is unavailable via API at this time)."
    )
    return await _get_vm_prices(location)


_DEFAULT_EVICTION_RATE = 60.0

# A bit of a hack to just hard-code these
_EXPECTED_MISSING_PRICES = {
    # These are burst instances, so they don't have spot versions
    ("Standard_B12ms", "spot"),
    ("Standard_B16ms", "spot"),
    ("Standard_B1ls", "spot"),
    ("Standard_B1ms", "spot"),
    ("Standard_B1s", "spot"),
    ("Standard_B20ms", "spot"),
    ("Standard_B2ms", "spot"),
    ("Standard_B2s", "spot"),
    ("Standard_B4ms", "spot"),
    ("Standard_B8ms", "spot"),
    # Not sure why these "Promo" SKUs don't have spot pricing
    ("Standard_DS11_v2_Promo", "spot"),
    ("Standard_DS12_v2_Promo", "spot"),
    ("Standard_DS13_v2_Promo", "spot"),
    ("Standard_DS14_v2_Promo", "spot"),
    ("Standard_DS2_v2_Promo", "spot"),
    ("Standard_DS3_v2_Promo", "spot"),
    ("Standard_DS4_v2_Promo", "spot"),
    ("Standard_DS5_v2_Promo", "spot"),
    # These SKUs just don't have any pricing, not sure why
    ("Standard_E96ias_v4", "on_demand"),
    ("Standard_E96ias_v4", "spot"),
    ("Standard_ND40s_v3", "on_demand"),
    ("Standard_ND40s_v3", "spot"),
    ("Standard_NV12s_v2", "on_demand"),
    ("Standard_NV12s_v2", "spot"),
    ("Standard_NV24s_v2", "on_demand"),
    ("Standard_NV24s_v2", "spot"),
    ("Standard_NV6s_v2", "on_demand"),
    ("Standard_NV6s_v2", "spot"),
}


async def get_vm_types(location: str) -> List[CloudInstanceType]:
    # TODO we should also take quotas into account, the default quotas will stop us from
    # creating a lot of VM sizes

    vm_skus = await _get_vm_skus(location)
    pricing_eviction_data = await _get_cached_vm_prices_and_eviction(location)

    results = []

    for name, consumable_resources in vm_skus:
        for on_demand_or_spot in ON_DEMAND_OR_SPOT_VALUES:
            price_eviction = pricing_eviction_data.get((name, on_demand_or_spot))
            if price_eviction is None:
                if (name, on_demand_or_spot) not in _EXPECTED_MISSING_PRICES:
                    print(
                        f"Warning skipping {name}, {on_demand_or_spot} because no "
                        "pricing/eviction rate data is available"
                    )
            else:
                price, eviction_rate = price_eviction
                if eviction_rate is None:
                    eviction_rate = _DEFAULT_EVICTION_RATE
                results.append(
                    CloudInstanceType(
                        name,
                        on_demand_or_spot,
                        price,
                        ResourcesInternal(
                            consumable_resources,
                            {EVICTION_RATE_INVERSE: 100 - eviction_rate},
                        ),
                    )
                )

    return results



"""
This is code that "should" be part of ec2_instance_allocation.py, but is also needed by
adjust_ec2_instances.

This module contains code that will run in the generic AWS Lambda environment, so it
should not refer to any outside code
"""
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)

if TYPE_CHECKING:
    from typing_extensions import Literal

import botocore.exceptions

_T = TypeVar("_T")

# a dynamodb table that holds information about EC2 instances we've created and what has
# been allocated to which instances
_EC2_ALLOC_TABLE_NAME = "meadowrun_ec2_alloc_table"

# names of attributes/keys in the EC2 alloc table
_PUBLIC_ADDRESS = "public_address"
_INSTANCE_ID = "instance_id"
_RESOURCES_AVAILABLE = "resources_available"
_NON_CONSUMABLE_RESOURCES = "non_consumable_resources"
_RESOURCES_ALLOCATED = "resources_allocated"
_ALLOCATED_TIME = "allocated_time"
_LAST_UPDATE_TIME = "last_update_time"
_PREVENT_FURTHER_ALLOCATION = "prevent_further_allocation"
_RUNNING_JOBS = "running_jobs"
_JOB_ID = "job_id"

# A tag for EC2 instances that are created using ec2_alloc
_EC2_ALLOC_TAG = "meadowrun_ec2_alloc"
_EC2_ALLOC_TAG_VALUE = "TRUE"

_MEADOWRUN_GENERATED_DOCKER_REPO = "meadowrun_generated"


@overload
def ignore_boto3_error_code(
    func: Callable[[], _T],
    error_codes: Union[str, Set[str]],
    return_error_code: Literal[False] = False,
) -> Union[Tuple[Literal[True], _T], Tuple[Literal[False], None]]:
    pass


@overload
def ignore_boto3_error_code(
    func: Callable[[], _T],
    error_codes: Union[str, Set[str]],
    return_error_code: Literal[True],
) -> Union[Tuple[Literal[True], _T, None], Tuple[Literal[False], None, str]]:
    pass


def ignore_boto3_error_code(
    func: Callable[[], _T],
    error_codes: Union[str, Set[str]],
    return_error_code: bool = False,
) -> Union[
    Tuple[Literal[True], _T],
    Tuple[Literal[False], None],
    Tuple[Literal[True], _T, None],
    Tuple[Literal[False], None, str],
]:
    """
    Calls func.
    - If func succeeds:
        - And return_error_code is False, returns (True, result of func)
        - And return_error_code is True, returns (True, result of func, None)
    - If func raises a boto3 error with the specified code:
        - And return_error_code is False, returns (False, None)
        - And return_error_code is True, returns (True, None, error code), where error
          code is one of the strings in the error_codes argument
    - If func raises any other type of exception (i.e. a boto3 error with a different
      code or a different type of error altogether), then the exception is raised
      normally.
    """
    if isinstance(error_codes, str):
        error_codes = {error_codes}

    try:
        result = func()
        if return_error_code:
            return True, result, None
        else:
            return True, result
    except botocore.exceptions.ClientError as e:
        if "Error" in e.response:
            error = e.response["Error"]
            if "Code" in error and error["Code"] in error_codes:
                if return_error_code:
                    return False, None, error["Code"]
                else:
                    return False, None

        raise


async def ignore_boto3_error_code_async(
    func: Awaitable[_T],
    error_codes: Union[str, Set[str]],
) -> Union[Tuple[Literal[True], _T, None], Tuple[Literal[False], None, str]]:
    """Same semantics as ignore_boto3_error_code"""
    if isinstance(error_codes, str):
        error_codes = {error_codes}

    try:
        result = await func
        return True, result, None
    except botocore.exceptions.ClientError as e:
        if "Error" in e.response:
            error = e.response["Error"]
            if "Code" in error and error["Code"] in error_codes:
                return False, None, error["Code"]

        raise



"""
This is a miniaturized/custom version of the Azure SDK for python
(https://github.com/Azure/azure-sdk-for-python). This should contain no
meadowrun-specific code.

We do this because the actual SDK requires dozens of packages that are not up to date on
conda-forge and take up ~500MB of disk space. Incidentally, we've also noticed
performance improvements by moving off of the Azure SDK.
"""



"""
This is very similar to run_job_local_main.py but instead of communicating with the
client via the local file system, it communicates via an S3 compatible storage bucket.
"""

import argparse
import asyncio
import logging
import os
import time
import traceback

import filelock

import meadowrun.run_job_local
from meadowrun.k8s_integration.storage_spec import (
    add_storage_spec_arguments_to_parser,
    storage_bucket_from_parsed_args,
)
from meadowrun.storage_keys import storage_key_job_to_run, storage_key_process_state
from meadowrun.meadowrun_pb2 import ProcessState, Job
from meadowrun.k8s_integration.is_job_running import _JOB_IS_RUNNING_FILE
from meadowrun.k8s_integration.k8s_main import _LAST_JOB_TIMESTAMP_FILE


async def main() -> None:
    with filelock.FileLock(_JOB_IS_RUNNING_FILE, 0):
        with open(_LAST_JOB_TIMESTAMP_FILE, "w", encoding="utf-8") as f:
            f.write(str(time.time()))

        logging.basicConfig(level=logging.INFO)

        # parse arguments

        parser = argparse.ArgumentParser()

        # arguments for finding the storage bucket (i.e. S3-compatible store) where we
        # can find inputs and put outputs
        parser.add_argument("--job-id", required=True)
        add_storage_spec_arguments_to_parser(parser)

        args = parser.parse_args()

        job_id: str = args.job_id
        suffix = os.environ.get(
            "MEADOWRUN_WORKER_INDEX", os.environ["JOB_COMPLETION_INDEX"]
        )

        # prepare storage bucket, filenames and pickle protocol for the result
        storage_bucket = await storage_bucket_from_parsed_args(args)
        async with storage_bucket:
            meadowrun.func_worker_storage_helper.FUNC_WORKER_STORAGE_BUCKET = (
                storage_bucket
            )

            # run the job
            try:
                job = Job.FromString(
                    await storage_bucket.get_bytes(storage_key_job_to_run(job_id))
                )
                # hypothetically we should make sure MEADOWRUN_STORAGE_USERNAME and
                # MEADOWRUN_STORAGE_PASSWORD should get added to the Job object because
                # if we are running _indexed_map_worker we will need those variables.
                # However, this only happens in the non-container job case, in which
                # case the environment variables should just get inherited by the child
                # process that we start.
                first_state, continuation = await meadowrun.run_job_local.run_local(
                    job, None, storage_bucket, compile_environment_in_container=False
                )
                # TODO to be analogous to run_job_local_main.py we should write
                # f"{storage_file_prefix}.initial_process_state to the storage bucket
                # here

                if (
                    first_state.state != ProcessState.ProcessStateEnum.RUNNING
                    or continuation is None
                ):
                    await storage_bucket.write_bytes(
                        first_state.SerializeToString(),
                        storage_key_process_state(job_id, suffix),
                    )
                else:
                    final_process_state = await continuation

                    await storage_bucket.write_bytes(
                        final_process_state.SerializeToString(),
                        storage_key_process_state(job_id, suffix),
                    )
            except:  # noqa: E722
                # so we know what's wrong
                traceback.print_exc()
                raise


if __name__ == "__main__":
    asyncio.run(main())



import datetime
import json
import os
import time
from typing import Any, Optional

from meadowrun._vendor.platformdirs import PlatformDirs


MEADOWRUN_DIRS = PlatformDirs(appname="meadowrun", appauthor=False)


def get_cached_json(name: str, freshness: datetime.timedelta) -> Optional[Any]:
    """Returns the cached JSON file from the standard cache dir,
    if it is not too old.

    Args:
        name (str): The name of the file. (not a path)
        freshness (timedelta): The maximum age of the file.

    Returns:
        Optional[Any]: The JSON data, or None if the file is too old.
    """
    file = os.path.join(MEADOWRUN_DIRS.user_cache_dir, name)
    if (
        not os.path.exists(file)
        or os.path.getmtime(file) + freshness.total_seconds() < time.time()
    ):
        return None

    with open(file, "r", encoding="utf-8") as f:
        return json.load(f)


def save_json_to_cache(name: str, json_data: Any) -> None:
    """Saves the JSON data to the standard cache dir.

    Args:
        name (str): The name of the file. (not a path)
        json_data (Any): The JSON data.
    """
    if not os.path.exists(MEADOWRUN_DIRS.user_cache_dir):
        os.makedirs(MEADOWRUN_DIRS.user_cache_dir, exist_ok=True)
    file = os.path.join(MEADOWRUN_DIRS.user_cache_dir, name)
    with open(file, "w", encoding="utf-8") as f:
        json.dump(json_data, f)


def clear_cache(name: str) -> None:
    try:
        os.remove(os.path.join(MEADOWRUN_DIRS.user_cache_dir, name))
    except OSError:
        pass



import binascii
import collections
import datetime
import enum
import hashlib
import io
from typing import Any
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING
from urllib.parse import quote

import rsa
from meadowrun._vendor.gcloud.aio.auth import (
    BUILD_GCLOUD_REST,
)  # pylint: disable=no-name-in-module
from meadowrun._vendor.gcloud.aio.auth import Token  # pylint: disable=no-name-in-module
from meadowrun._vendor.gcloud.aio.storage.constants import DEFAULT_TIMEOUT
from pyasn1.codec.der import decoder
from pyasn1_modules import pem
from pyasn1_modules.rfc5208 import PrivateKeyInfo

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]

if TYPE_CHECKING:
    from .bucket import Bucket  # pylint: disable=cyclic-import


HOST = "storage.googleapis.com"

PKCS1_MARKER = ("-----BEGIN RSA PRIVATE KEY-----", "-----END RSA PRIVATE KEY-----")
PKCS8_MARKER = ("-----BEGIN PRIVATE KEY-----", "-----END PRIVATE KEY-----")
PKCS8_SPEC = PrivateKeyInfo()


class PemKind(enum.Enum):
    """
    Tracks the response of ``pem.readPemBlocksFromFile(key, *args)``>

    Note that the specified method returns ``(marker_id, key_bytes)``, where
    ``marker_id`` is the integer index of the matching ``arg`` (or -1 if no
    match was found.

    For example::

        (marker_id, _) = pem.readPemBlocksFromFile(key, PKCS1_MARKER,
                                                   PCKS8_MARKER)
        if marker_id == -1:
            # "key" did not match either type or was invalid
        if marker_id == 0:
            # "key" matched the zeroth provided marker arg, eg. PKCS1_MARKER
        if marker_id == 1:
            # "key" matched the zeroth provided marker arg, eg. PKCS8_MARKER
    """

    INVALID = -1
    PKCS1 = 0
    PKCS8 = 1


class Blob:
    def __init__(self, bucket: "Bucket", name: str, metadata: Dict[str, Any]) -> None:
        self.__dict__.update(**metadata)

        self.bucket = bucket
        self.name = name
        self.size: int = int(self.size)

    @property
    def chunk_size(self) -> int:
        return self.size + (262144 - (self.size % 262144))

    async def download(
        self, timeout: int = DEFAULT_TIMEOUT, session: Optional[Session] = None
    ) -> Any:
        return await self.bucket.storage.download(
            self.bucket.name, self.name, timeout=timeout, session=session
        )

    async def upload(
        self, data: Any, session: Optional[Session] = None
    ) -> Dict[str, Any]:
        metadata = await self.bucket.storage.upload(
            self.bucket.name, self.name, data, session=session
        )

        self.__dict__.update(metadata)
        return metadata

    async def get_signed_url(  # pylint: disable=too-many-locals
        self,
        expiration: int,
        headers: Optional[Dict[str, str]] = None,
        query_params: Optional[Dict[str, Any]] = None,
        http_method: str = "GET",
        token: Optional[Token] = None,
    ) -> str:
        """
        Create a temporary access URL for Storage Blob accessible by anyone
        with the link.

        Adapted from Google Documentation:
        https://cloud.google.com/storage/docs/access-control/signing-urls-manually#python-sample
        """
        if expiration > 604800:
            raise ValueError(
                "expiration time can't be longer than 604800 " "seconds (7 days)"
            )

        quoted_name = quote(self.name, safe=b"/~")
        canonical_uri = f"/{quoted_name}"

        datetime_now = datetime.datetime.utcnow()
        request_timestamp = datetime_now.strftime("%Y%m%dT%H%M%SZ")
        datestamp = datetime_now.strftime("%Y%m%d")

        # TODO: support for GCE_METADATA and AUTHORIZED_USER tokens. It should
        # be possible via API, but seems to not be working for us in some
        # cases.
        #
        # https://cloud.google.com/iam/docs/reference/credentials/rest/v1/
        # projects.serviceAccounts/signBlob
        token = token or self.bucket.storage.token
        client_email = token.service_data.get("client_email")
        private_key = token.service_data.get("private_key")
        if not client_email or not private_key:
            raise KeyError(
                "Blob signing is only suported for tokens with "
                "explicit client_email and private_key data; "
                "please check your token points to a JSON service "
                "account file"
            )

        credential_scope = f"{datestamp}/auto/storage/goog4_request"
        credential = f"{client_email}/{credential_scope}"

        headers = headers or {}
        headers["host"] = f"{self.bucket.name}.{HOST}"

        ordered_headers = collections.OrderedDict(sorted(headers.items()))
        canonical_headers = "".join(
            f"{str(k).lower()}:{str(v).lower()}\n" for k, v in ordered_headers.items()
        )

        signed_headers = ";".join(f"{str(k).lower()}" for k in ordered_headers.keys())

        query_params = query_params or {}
        query_params["X-Goog-Algorithm"] = "GOOG4-RSA-SHA256"
        query_params["X-Goog-Credential"] = credential
        query_params["X-Goog-Date"] = request_timestamp
        query_params["X-Goog-Expires"] = expiration
        query_params["X-Goog-SignedHeaders"] = signed_headers

        ordered_query_params = collections.OrderedDict(sorted(query_params.items()))

        canonical_query_str = "&".join(
            f'{quote(str(k), safe="")}={quote(str(v), safe="")}'
            for k, v in ordered_query_params.items()
        )

        canonical_req = "\n".join(
            [
                http_method,
                canonical_uri,
                canonical_query_str,
                canonical_headers,
                signed_headers,
                "UNSIGNED-PAYLOAD",
            ]
        )
        canonical_req_hash = hashlib.sha256(canonical_req.encode()).hexdigest()

        str_to_sign = "\n".join(
            [
                "GOOG4-RSA-SHA256",
                request_timestamp,
                credential_scope,
                canonical_req_hash,
            ]
        )

        # N.B. see the ``PemKind`` enum
        marker_id, key_bytes = pem.readPemBlocksFromFile(
            io.StringIO(private_key), PKCS1_MARKER, PKCS8_MARKER
        )
        if marker_id == PemKind.INVALID.value:
            raise ValueError("private key is invalid or unsupported")

        if marker_id == PemKind.PKCS8.value:
            # convert from pkcs8 to pkcs1
            key_info, remaining = decoder.decode(key_bytes, asn1Spec=PKCS8_SPEC)
            if remaining != b"":
                raise ValueError(
                    "could not read PKCS8 key: found extra bytes", remaining
                )

            private_key_info = key_info.getComponentByName("privateKey")
            key_bytes = private_key_info.asOctets()

        key = rsa.key.PrivateKey.load_pkcs1(key_bytes, format="DER")
        signed_blob = rsa.pkcs1.sign(str_to_sign.encode(), key, "SHA-256")

        signature = binascii.hexlify(signed_blob).decode()

        return (
            f"https://{self.bucket.name}.{HOST}{canonical_uri}?"
            f"{canonical_query_str}&X-Goog-Signature={signature}"
        )



import asyncio
import sys
import traceback
import warnings
from typing import Any

if sys.platform == "win32":
    # The default event loop type was changed from the selector event loop to the
    # proactor event loop in python 3.8. To make meadowrun work on python 3.7, we need
    # to use proactor event loops--aiohttp and asyncio.subprocess don't seem to work
    # with the selector event loop
    if sys.version_info < (3, 8):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    # There's a bug in the proactor event loop that causes messages like "Exception
    # ignored in: # <function _ProactorBasePipeTransport.__del__ at 0x...>" to be
    # printed. We make extensive use of aiohttp, so we produce lots of messages like
    # this. Below is monkey-patching the fix for this issue, see
    # https://github.com/python/cpython/pull/92842/files via
    # https://github.com/python/cpython/issues/83413 and
    # https://github.com/aio-libs/aiohttp/issues/4324 We should keep an eye on that
    # change and refine this version check as the fix gets backported to older versions
    # of python
    if sys.version_info < (3, 12):
        try:
            from asyncio.proactor_events import _ProactorBasePipeTransport

            def new__del__(self: Any, _warn: Any = warnings.warn) -> None:
                if self._sock is not None:
                    _warn(f"unclosed transport {self!r}", ResourceWarning, source=self)
                    self._sock.close()

            # we can't just do a simple assignment because that causes a mypy error as
            # described in https://github.com/python/mypy/issues/2427 and we can't just
            # "type: ignore" that error because we will get a "type: ignore unused"
            # error on Linux (even though that seems like it should work based on
            # https://github.com/python/mypy/issues/8547 and
            # https://github.com/python/mypy/issues/8823). Luckily the setattr
            # workaround works.
            setattr(_ProactorBasePipeTransport, "__del__", new__del__)
        except Exception:
            traceback.print_exc()


from meadowrun.aws_integration.ec2_instance_allocation import AllocEC2Instance
from meadowrun.azure_integration.azure_instance_allocation import AllocAzureVM
from meadowrun.deployment_spec import (
    AwsSecret,
    AzureSecret,
    CondaEnvironmentYmlFile,
    ContainerAtDigestInterpreter,
    ContainerInterpreter,
    ContainerInterpreterBase,
    Deployment,
    InterpreterSpecFile,
    KubernetesSecret,
    LocalCondaInterpreter,
    LocalInterpreter,
    LocalPipInterpreter,
    PipRequirementsFile,
    PoetryProjectPath,
    PreinstalledInterpreter,
    Secret,
)
from meadowrun.k8s_integration.k8s import Kubernetes, StorageBucketSpec
from meadowrun.k8s_integration.storage_spec import (
    GenericStorageBucketSpec,
    GoogleBucketSpec,
)
from meadowrun.run_job import (
    TaskResult,
    run_command,
    run_function,
    run_map,
    run_map_as_completed,
)
from meadowrun.alloc_cloud_instance import AllocCloudInstance
from meadowrun.run_job_core import (
    AllocVM,
    Resources,
    RunMapTasksFailedException,
    SshHost,
    TaskException,
)
from meadowrun.run_job_core import Host
from meadowrun.run_job_local import MACHINE_CACHE_FOLDER
from meadowrun.version import __version__

__all__ = [
    "__version__",
    "AllocVM",
    "AllocCloudInstance",
    "AllocEC2Instance",
    "AllocAzureVM",
    "run_command",
    "run_function",
    "run_map",
    "run_map_as_completed",
    "AwsSecret",
    "AzureSecret",
    "CondaEnvironmentYmlFile",
    "ContainerAtDigestInterpreter",
    "ContainerInterpreter",
    "ContainerInterpreterBase",
    "Deployment",
    "InterpreterSpecFile",
    "KubernetesSecret",
    "LocalCondaInterpreter",
    "LocalInterpreter",
    "LocalPipInterpreter",
    "PipRequirementsFile",
    "PoetryProjectPath",
    "PreinstalledInterpreter",
    "Secret",
    "Host",
    "MACHINE_CACHE_FOLDER",
    "Resources",
    "SshHost",
    "Kubernetes",
    "StorageBucketSpec",
    "GenericStorageBucketSpec",
    "GoogleBucketSpec",
    "RunMapTasksFailedException",
    "TaskResult",
    "TaskException",
]



from __future__ import annotations

import argparse
import asyncio
import asyncio.subprocess
import datetime
import logging
from typing import TYPE_CHECKING, Optional, Dict, Any, ItemsView, Union, Iterable, Tuple

import psutil

from meadowrun.aws_integration.aws_core import (
    _get_default_region_name,
    _get_ec2_metadata,
)
from meadowrun.aws_integration.ec2_instance_allocation import EC2InstanceRegistrar
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import _ALLOCATED_TIME
from meadowrun.azure_integration.azure_meadowrun_core import (
    get_default_location,
    get_current_vm_name,
)
from meadowrun.azure_integration.azure_instance_allocation import AzureInstanceRegistrar

if TYPE_CHECKING:
    from meadowrun.instance_allocation import InstanceRegistrar
from meadowrun.run_job_core import CloudProvider, CloudProviderType

# If a job is allocated but we never see a pid file for it, we assume after this amount
# of time that the client process crashed
_ALLOCATED_BUT_NOT_RUNNING_TIMEOUT = datetime.timedelta(minutes=7)


def _try_read_pid_file(job_id: str) -> Optional[int]:
    """
    Tries to get the pid file that run_one_job_main.py will create for a particular
    job_id. If the pid file does not exist or does not contain a valid integer, returns
    None.
    """
    try:
        f = open(f"/var/meadowrun/io/{job_id}.pid", "r", encoding="utf-8")
    except FileNotFoundError:
        return None

    try:
        pid_str = f.read()
        try:
            return int(pid_str)
        except ValueError:
            print(
                f"Warning, pid file for {job_id} could not be interpreted as an int: "
                f"{pid_str}"
            )
            return None
    finally:
        f.close()


async def async_main(
    cloud: CloudProviderType,
    cloud_region_name: str,
    job_id: Optional[str],
    allocated_but_not_running_timeout: datetime.timedelta,
) -> None:
    """
    Deallocates job(s) from the EC2 alloc table. public_address needs to correspond to
    the machine that we're currently running on. If job_id is not specified, we check
    the EC2 alloc table for all jobs currently allocated on this machine. If any of them
    are complete or never ran successfully, we deallocate them. If job_id is specified,
    we only check whether we need to deallocate the specified job.
    """
    now = datetime.datetime.utcnow()

    if cloud == "EC2":
        instance_name = await _get_ec2_metadata("instance-id")
        if cloud_region_name == "default":
            cloud_region_name = await _get_default_region_name()
        instance_registrar: InstanceRegistrar = EC2InstanceRegistrar(
            cloud_region_name, "raise"
        )
    elif cloud == "AzureVM":
        instance_name = await get_current_vm_name()
        if cloud_region_name == "default":
            cloud_region_name = get_default_location()
        instance_registrar = AzureInstanceRegistrar(cloud_region_name, "raise")
    else:
        raise ValueError(f"Unexpected value for cloud_provider: {cloud}")

    async with instance_registrar:
        if not instance_name:
            raise ValueError(
                "Cannot deallocate jobs because we can't get the name/id of the "
                f"current {cloud} instance (maybe we're not running on a {cloud} "
                "instance?)"
            )

        registered_instance = await instance_registrar.get_registered_instance(
            instance_name
        )
        if job_id:
            job = registered_instance.get_running_jobs().get(job_id)
            if job is None:
                print(
                    f"job_id {job_id} was specified, but it does not exist in the EC2 "
                    "alloc table"
                )
                return
            jobs_to_deallocate: Union[
                Iterable[Tuple[str, Dict[str, Any]]], ItemsView[str, Dict[str, Any]]
            ] = [(job_id, job)]
        else:
            jobs_to_deallocate = registered_instance.get_running_jobs().items()

        for job_id, job in jobs_to_deallocate:
            pid = _try_read_pid_file(job_id)
            if pid is not None:
                if not psutil.pid_exists(pid):
                    print(
                        f"Deallocating {job_id} as the process {pid} is no longer "
                        "running"
                    )
                    await instance_registrar.deallocate_job_from_instance(
                        registered_instance, job_id
                    )
                else:
                    print(
                        f"Not deallocating {job_id} as the process {pid} is still "
                        "running"
                    )
            else:
                job_allocated_time = datetime.datetime.fromisoformat(
                    job[_ALLOCATED_TIME]
                )
                if now - job_allocated_time > allocated_but_not_running_timeout:
                    print(
                        "Warning: deallocating job that looks like it never ran "
                        f"{job_id} was allocated at {job_allocated_time}"
                    )
                    await instance_registrar.deallocate_job_from_instance(
                        registered_instance, job_id
                    )
                else:
                    print(
                        f"Not deallocating {job_id}: we don't have a pid file but it "
                        f"was allocated recently at {job_allocated_time}"
                    )


def main(
    cloud: CloudProviderType,
    cloud_region_name: str,
    job_id: Optional[str],
    allocated_but_not_running_timeout: datetime.timedelta,
) -> None:
    asyncio.run(
        async_main(
            cloud,
            cloud_region_name,
            job_id,
            allocated_but_not_running_timeout,
        )
    )


def command_line_main() -> None:
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("--cloud", required=True, choices=CloudProvider)
    parser.add_argument("--cloud-region-name", required=True)
    parser.add_argument("--job-id")
    parser.add_argument("--allocated-but-not-running-timeout-seconds", type=int)
    args = parser.parse_args()

    if args.allocated_but_not_running_timeout_seconds is not None:
        allocated_but_not_running_timeout = datetime.timedelta(
            seconds=args.allocated_but_not_running_timeout_seconds
        )
    else:
        allocated_but_not_running_timeout = _ALLOCATED_BUT_NOT_RUNNING_TIMEOUT

    main(
        args.cloud,
        args.cloud_region_name,
        args.job_id,
        allocated_but_not_running_timeout,
    )


if __name__ == "__main__":
    command_line_main()



# TODO turn this into a real module that allows the user to specify configurations
import enum
import string
from typing import Optional

JOB_ID_VALID_CHARACTERS = set(string.ascii_letters + string.digits + "-_.")


# A placeholder for the same interpreter that meadowrun.agent is using. Mostly for
# testing, not recommended for normal use.
MEADOWRUN_INTERPRETER = "__MEADOWRUN_INTERPRETER__"


# Paths for meadowrun to bind to inside containers
MEADOWRUN_CODE_MOUNT_LINUX = "/meadowrun/code"
MEADOWRUN_IO_MOUNT_LINUX = "/meadowrun/io"


# names of resources and default values
MEMORY_GB = "memory_gb"
LOGICAL_CPU = "logical_cpu"
GPU = "gpu"
GPU_MEMORY = "gpu_memory"
# in the resources, we set 100 - eviction rate as a non-consumable resource
EVICTION_RATE_INVERSE = "eviction_rate_inverse"
INTEL = "intel"
AMD = "amd"
AVX = "avx"
INTEL_DEEP_LEARNING_BOOST = "intel_deep_learning_boost"


class AvxVersion(enum.IntEnum):
    NONE = 0
    AVX = 1
    AVX2 = 2
    AVX512 = 3


def avx_from_string(s: str) -> Optional[AvxVersion]:
    s = s.upper()
    if s != "NONE":
        try:
            return AvxVersion[s]
        except Exception:
            pass

    return None


# these are totally arbitrary values...
DEFAULT_MEMORY_GB_REQUIRED = 2
DEFAULT_LOGICAL_CPU_REQUIRED = 1


DEFAULT_PRIORITY = 100


# names of environment variables to communicate with the child process

# Will be set in child processes launched by an agent, gives the pid of the agent that
# launched the child process. Mostly for testing/debugging.
MEADOWRUN_AGENT_PID = "MEADOWRUN_AGENT_PID"


# specifies how often EC2 prices should get updated
EC2_PRICES_UPDATE_SECS = 60 * 30  # 30 minutes
